import equinox as eqx
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from esm.utils.structure.affine3d import (
    Affine3D as TorchAffine3D,
)
from esm.utils.structure.affine3d import (
    RotationMatrix as TorchRotationMatrix,
)
from esm.utils.structure.affine3d import (
    RotationQuat as TorchRotationQuat,
)
from esm.utils.structure.affine3d import (
    _graham_schmidt as torch_graham_schmidt,
)
from esm.utils.structure.affine3d import (
    _quat_invert as torch_quat_invert,
)
from esm.utils.structure.affine3d import (
    _quat_mult as torch_quat_mult,
)
from esm.utils.structure.affine3d import (
    _quat_rotation as torch_quat_rotation,
)
from esm.utils.structure.affine3d import (
    _sqrt_subgradient as torch_sqrt_subgradient,
)
from esm.utils.structure.affine3d import (
    build_affine3d_from_coordinates as torch_build_affine,
)

from jaxonmodels.models.esm import (
    Affine3D as JaxAffine3D,
)
from jaxonmodels.models.esm import (
    RotationMatrix as JaxRotationMatrix,
)
from jaxonmodels.models.esm import (
    RotationQuat as JaxRotationQuat,
)
from jaxonmodels.models.esm import (
    _graham_schmidt as jax_graham_schmidt,
)
from jaxonmodels.models.esm import (
    _quat_invert as jax_quat_invert,
)
from jaxonmodels.models.esm import (
    _quat_mult as jax_quat_mult,
)
from jaxonmodels.models.esm import (
    _quat_rotation as jax_quat_rotation,
)
from jaxonmodels.models.esm import (
    _sqrt_subgradient as jax_sqrt_subgradient,
)
from jaxonmodels.models.esm import build_affine3d_from_coordinates as jax_build_affine


def _rand(*shape, seed=42):
    np.random.seed(seed)
    return np.random.randn(*shape).astype(np.float32)


def test_quat_mult():
    a = _rand(5, 4)
    b = _rand(5, 4, seed=43)
    torch_out = torch_quat_mult(torch.from_numpy(a), torch.from_numpy(b)).numpy()
    jax_out = np.array(jax_quat_mult(jnp.array(a), jnp.array(b)))
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_quat_mult_batched():
    a = _rand(3, 5, 4)
    b = _rand(3, 5, 4, seed=43)
    torch_out = torch_quat_mult(torch.from_numpy(a), torch.from_numpy(b)).numpy()
    jax_out = np.array(jax_quat_mult(jnp.array(a), jnp.array(b)))
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_quat_invert():
    q = _rand(5, 4)
    torch_out = torch_quat_invert(torch.from_numpy(q)).numpy()
    jax_out = np.array(jax_quat_invert(jnp.array(q)))
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_quat_rotation():
    q = _rand(5, 4)
    p = _rand(5, 3, seed=43)
    torch_out = torch_quat_rotation(torch.from_numpy(q), torch.from_numpy(p)).numpy()
    jax_out = np.array(jax_quat_rotation(jnp.array(q), jnp.array(p)))
    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_sqrt_subgradient():
    x = np.array([-1.0, 0.0, 0.5, 2.0, -0.3], dtype=np.float32)
    torch_out = torch_sqrt_subgradient(torch.from_numpy(x)).numpy()
    jax_out = np.array(jax_sqrt_subgradient(jnp.array(x)))
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_graham_schmidt():
    x_axis = _rand(5, 3)
    xy_plane = _rand(5, 3, seed=43)
    torch_out = torch_graham_schmidt(
        torch.from_numpy(x_axis), torch.from_numpy(xy_plane)
    ).numpy()
    jax_out = np.array(jax_graham_schmidt(jnp.array(x_axis), jnp.array(xy_plane)))
    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_rotation_quat_identity():
    shape = (3, 5)
    torch_out = TorchRotationQuat.identity(shape).tensor.numpy()
    jax_out = np.array(JaxRotationQuat.identity(shape).tensor)
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_quat_compose():
    a = _rand(5, 4)
    b = _rand(5, 4, seed=43)
    torch_out = (
        TorchRotationQuat(torch.from_numpy(a))
        .compose(TorchRotationQuat(torch.from_numpy(b)))
        .tensor.numpy()
    )
    jax_out = np.array(
        JaxRotationQuat(jnp.array(a)).compose(JaxRotationQuat(jnp.array(b))).tensor
    )
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_quat_apply():
    q = _rand(5, 4)
    p = _rand(5, 3, seed=43)
    torch_out = (
        TorchRotationQuat(torch.from_numpy(q), normalized=True)
        .apply(torch.from_numpy(p))
        .numpy()
    )
    jax_out = np.array(
        JaxRotationQuat(jnp.array(q), normalized=True).apply(jnp.array(p))
    )
    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_rotation_quat_invert():
    q = _rand(5, 4)
    torch_out = TorchRotationQuat(torch.from_numpy(q)).invert().tensor.numpy()
    jax_out = np.array(JaxRotationQuat(jnp.array(q)).invert().tensor)
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_quat_to_matrix():
    q = _rand(5, 4)
    torch_out = (
        TorchRotationQuat(torch.from_numpy(q), normalized=True)
        .as_matrix()
        .to_3x3()
        .numpy()
    )
    jax_out = np.array(
        JaxRotationQuat(jnp.array(q), normalized=True).as_matrix().to_3x3()
    )
    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_rotation_matrix_identity():
    shape = (3, 5)
    torch_out = TorchRotationMatrix.identity(shape).to_3x3().numpy()
    jax_out = np.array(JaxRotationMatrix.identity(shape).to_3x3())
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_matrix_compose():
    r1 = _rand(5, 3, 3)
    r2 = _rand(5, 3, 3, seed=43)
    torch_result = TorchRotationMatrix(torch.from_numpy(r1)).compose(
        TorchRotationMatrix(torch.from_numpy(r2))
    )
    jax_result = JaxRotationMatrix(jnp.array(r1)).compose(
        JaxRotationMatrix(jnp.array(r2))
    )
    assert np.allclose(
        torch_result.tensor.numpy(), np.array(jax_result.tensor), atol=1e-5
    )


def test_rotation_matrix_apply():
    r = _rand(5, 3, 3)
    p = _rand(5, 3, seed=43)
    torch_out = (
        TorchRotationMatrix(torch.from_numpy(r)).apply(torch.from_numpy(p)).numpy()
    )
    jax_out = np.array(JaxRotationMatrix(jnp.array(r)).apply(jnp.array(p)))
    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_rotation_matrix_invert():
    r = _rand(5, 3, 3)
    torch_out = TorchRotationMatrix(torch.from_numpy(r)).invert().tensor.numpy()
    jax_out = np.array(JaxRotationMatrix(jnp.array(r)).invert().tensor)
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_matrix_to_quat():
    q = _rand(5, 4)
    torch_mat = TorchRotationQuat(torch.from_numpy(q), normalized=True).as_matrix()
    torch_out = torch_mat.as_quat().tensor.numpy()
    jax_mat = JaxRotationQuat(jnp.array(q), normalized=True).as_matrix()
    jax_out = np.array(jax_mat.as_quat().tensor)
    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("shape", [(5,), (3, 5), (2, 3, 4)])
def test_affine3d_compose(shape):
    t1 = _rand(*shape, 3)
    r1 = _rand(*shape, 3, 3)
    t2 = _rand(*shape, 3, seed=43)
    r2 = _rand(*shape, 3, 3, seed=44)

    torch_a1 = TorchAffine3D(
        torch.from_numpy(t1), TorchRotationMatrix(torch.from_numpy(r1))
    )
    torch_a2 = TorchAffine3D(
        torch.from_numpy(t2), TorchRotationMatrix(torch.from_numpy(r2))
    )
    torch_out = torch_a1.compose(torch_a2).tensor.numpy()

    jax_a1 = JaxAffine3D(jnp.array(t1), JaxRotationMatrix(jnp.array(r1)))
    jax_a2 = JaxAffine3D(jnp.array(t2), JaxRotationMatrix(jnp.array(r2)))
    jax_out = np.array(jax_a1.compose(jax_a2).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("shape", [(5,), (3, 5), (2, 3, 4)])
def test_affine3d_apply(shape):
    t = _rand(*shape, 3)
    r = _rand(*shape, 3, 3)
    p = _rand(*shape, 3, seed=43)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.apply(torch.from_numpy(p)).numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.apply(jnp.array(p)))

    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("shape", [(5,), (3, 5), (2, 3, 4)])
def test_affine3d_invert(shape):
    t = _rand(*shape, 3)
    r = _rand(*shape, 3, 3)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.invert().tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.invert().tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("last_dim", [7, 12])
def test_affine3d_from_tensor(last_dim):
    if last_dim == 7:
        t = _rand(5, 7)
    else:
        t = _rand(5, 12)

    torch_a = TorchAffine3D.from_tensor(torch.from_numpy(t))
    torch_out = torch_a.tensor.numpy()

    jax_a = JaxAffine3D.from_tensor(jnp.array(t))
    jax_out = np.array(jax_a.tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_affine3d_from_graham_schmidt():
    neg_x = _rand(5, 3)
    origin = _rand(5, 3, seed=43)
    xy = _rand(5, 3, seed=44)

    torch_a = TorchAffine3D.from_graham_schmidt(
        torch.from_numpy(neg_x), torch.from_numpy(origin), torch.from_numpy(xy)
    )
    torch_out = torch_a.tensor.numpy()

    jax_a = JaxAffine3D.from_graham_schmidt(
        jnp.array(neg_x), jnp.array(origin), jnp.array(xy)
    )
    jax_out = np.array(jax_a.tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("shape", [(5,), (3, 5)])
def test_affine3d_scale(shape):
    t = _rand(*shape, 3)
    r = _rand(*shape, 3, 3)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.scale(2.5).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.scale(2.5).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_affine3d_cat():
    t1 = _rand(5, 3)
    r1 = _rand(5, 3, 3)
    t2 = _rand(3, 3, seed=43)
    r2 = _rand(3, 3, 3, seed=44)

    torch_a1 = TorchAffine3D(
        torch.from_numpy(t1), TorchRotationMatrix(torch.from_numpy(r1))
    )
    torch_a2 = TorchAffine3D(
        torch.from_numpy(t2), TorchRotationMatrix(torch.from_numpy(r2))
    )
    torch_out = TorchAffine3D.cat([torch_a1, torch_a2], dim=0).tensor.numpy()

    jax_a1 = JaxAffine3D(jnp.array(t1), JaxRotationMatrix(jnp.array(r1)))
    jax_a2 = JaxAffine3D(jnp.array(t2), JaxRotationMatrix(jnp.array(r2)))
    jax_out = np.array(JaxAffine3D.cat([jax_a1, jax_a2], dim=0).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_affine3d_from_tensor_case6():
    t = _rand(5, 6)

    torch_a = TorchAffine3D.from_tensor(torch.from_numpy(t))
    torch_out = torch_a.tensor.numpy()

    jax_a = JaxAffine3D.from_tensor(jnp.array(t))
    jax_out = np.array(jax_a.tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


@pytest.mark.parametrize("shape", [(5,), (3, 5)])
def test_affine3d_identity_from_shape(shape):
    torch_out = TorchAffine3D.identity(shape).tensor.numpy()
    jax_out = np.array(JaxAffine3D.identity(shape).tensor)
    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_affine3d_identity_from_affine():
    t = _rand(5, 3)
    r = _rand(5, 3, 3)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = TorchAffine3D.identity(torch_a).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(JaxAffine3D.identity(jax_a).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


@pytest.mark.parametrize("shape", [(5,), (3, 5)])
def test_affine3d_mask_with_zero(shape):
    t = _rand(*shape, 3)
    r = _rand(*shape, 3, 3)
    np.random.seed(99)
    m = np.random.choice([True, False], size=shape)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.mask(torch.from_numpy(m), with_zero=True).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.mask(jnp.array(m), with_zero=True).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


@pytest.mark.parametrize("shape", [(5,), (3, 5)])
def test_affine3d_mask_with_identity(shape):
    t = _rand(*shape, 3)
    r = _rand(*shape, 3, 3)
    np.random.seed(99)
    m = np.random.choice([True, False], size=shape)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.mask(torch.from_numpy(m), with_zero=False).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.mask(jnp.array(m), with_zero=False).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_convert_compose_quat_to_matrix():
    q = _rand(5, 4)
    r = _rand(5, 3, 3, seed=43)

    torch_rm = TorchRotationMatrix(torch.from_numpy(r))
    torch_rq = TorchRotationQuat(torch.from_numpy(q))
    torch_out = torch_rm.convert_compose(torch_rq).tensor.numpy()

    jax_rm = JaxRotationMatrix(jnp.array(r))
    jax_rq = JaxRotationQuat(jnp.array(q))
    jax_out = np.array(jax_rm.convert_compose(jax_rq).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_affine3d_compose_rotation():
    t = _rand(5, 3)
    r1 = _rand(5, 3, 3)
    r2 = _rand(5, 3, 3, seed=43)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r1))
    )
    torch_out = torch_a.compose_rotation(
        TorchRotationMatrix(torch.from_numpy(r2))
    ).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r1)))
    jax_out = np.array(jax_a.compose_rotation(JaxRotationMatrix(jnp.array(r2))).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


@pytest.mark.parametrize("shape", [(5,), (2, 3, 4)])
def test_rotation_matrix_to_quat_roundtrip(shape):
    q = _rand(*shape, 4)
    torch_rq = TorchRotationQuat(torch.from_numpy(q), normalized=True)
    torch_out = torch_rq.as_matrix().as_quat().tensor.numpy()

    jax_rq = JaxRotationQuat(jnp.array(q), normalized=True)
    jax_out = np.array(jax_rq.as_matrix().as_quat().tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_affine3d_getitem():
    t = _rand(5, 3)
    r = _rand(5, 3, 3)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a[2].tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a[2].tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_quat_tensor_apply():
    q = _rand(5, 4)
    func = lambda x: x * 2.0 + 1.0

    torch_out = TorchRotationQuat(torch.from_numpy(q)).tensor_apply(func).tensor.numpy()
    jax_out = np.array(JaxRotationQuat(jnp.array(q)).tensor_apply(func).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_matrix_tensor_apply():
    r = _rand(5, 3, 3)
    func = lambda x: x * 0.5

    torch_out = (
        TorchRotationMatrix(torch.from_numpy(r)).tensor_apply(func).tensor.numpy()
    )
    jax_out = np.array(JaxRotationMatrix(jnp.array(r)).tensor_apply(func).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_affine3d_tensor_apply():
    t = _rand(5, 3)
    r = _rand(5, 3, 3)
    func = lambda x: x + 1.0

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.tensor_apply(func).tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.tensor_apply(func).tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_affine3d_as_matrix():
    q = _rand(5, 4)
    t = _rand(5, 3, seed=43)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationQuat(torch.from_numpy(q), normalized=True)
    )
    torch_out = torch_a.as_matrix().tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationQuat(jnp.array(q), normalized=True))
    jax_out = np.array(jax_a.as_matrix().tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_affine3d_as_quat():
    r = _rand(5, 3, 3)
    t = _rand(5, 3, seed=43)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a.as_quat().tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a.as_quat().tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-5)


def test_affine3d_getitem_none():
    t = _rand(5, 3)
    r = _rand(5, 3, 3)

    torch_a = TorchAffine3D(
        torch.from_numpy(t), TorchRotationMatrix(torch.from_numpy(r))
    )
    torch_out = torch_a[None].tensor.numpy()

    jax_a = JaxAffine3D(jnp.array(t), JaxRotationMatrix(jnp.array(r)))
    jax_out = np.array(jax_a[None].tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


def test_rotation_matrix_getitem_multidim():
    r = _rand(3, 5, 3, 3)

    torch_out = TorchRotationMatrix(torch.from_numpy(r))[1, 2:4].tensor.numpy()
    jax_out = np.array(JaxRotationMatrix(jnp.array(r))[1, 2:4].tensor)

    assert np.allclose(torch_out, jax_out, atol=1e-6)


@pytest.mark.parametrize(
    "seq_len, nan_ratio",
    [
        (7, 0.0),
        (11, 0.3),
        (11, 1.0),
    ],
)
def test_build_affine3d_from_coordinates(seq_len, nan_ratio):
    batch = 2
    np.random.seed(42)

    coords = np.random.randn(batch, seq_len, 3, 3).astype(np.float32)

    num_nan = int(seq_len * nan_ratio)
    if num_nan > 0:
        coords[:, :num_nan, :, :] = np.nan

    torch_coords = torch.from_numpy(coords.copy())
    torch_affine, torch_mask = torch_build_affine(torch_coords)
    torch_tensor = torch_affine.tensor.detach().numpy()
    torch_mask = torch_mask.numpy()

    jax_coords = jnp.array(coords)

    def call_single(c):
        affine, mask = jax_build_affine(c)
        return affine.tensor, mask

    jax_tensor, jax_mask = eqx.filter_vmap(call_single)(jax_coords)

    assert np.array_equal(torch_mask, np.array(jax_mask))
    assert np.allclose(torch_tensor, np.array(jax_tensor), atol=1e-5, equal_nan=True)



import math
from typing import Optional

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any
from jaxonlayers.functions import default_floating_dtype, selective_scan
from jaxtyping import Array, Float, Int, PRNGKeyArray


class SelectiveStateSpaceModel(eqx.Module):
    input_proj: eqx.nn.Linear
    delta_proj: eqx.nn.Linear
    A_log: Float[Array, "d_inner d_state"]
    D: Float[Array, " d_inner"]

    d_inner: int = eqx.field(static=True)
    dt_rank: int = eqx.field(static=True)
    d_state: int = eqx.field(static=True)

    def __init__(
        self,
        d_inner: int,
        dt_rank: int,
        d_state: int,
        use_input_proj_bias: bool = False,
        use_delta_proj_bias: bool = False,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.d_inner = d_inner
        self.dt_rank = dt_rank
        self.d_state = d_state
        (
            key,
            input_proj_key,
            delta_proj_key,
        ) = jax.random.split(key, 3)
        self.input_proj = eqx.nn.Linear(
            d_inner,
            dt_rank + d_state * 2,
            use_bias=use_input_proj_bias,
            key=input_proj_key,
            dtype=dtype,
        )

        self.delta_proj = eqx.nn.Linear(
            dt_rank,
            d_inner,
            use_bias=use_delta_proj_bias,
            key=delta_proj_key,
            dtype=dtype,
        )
        A = jnp.repeat(jnp.arange(1, d_state + 1), d_inner).reshape(d_inner, d_state)
        self.A_log = jnp.log(A)
        self.D = jnp.ones(d_inner, dtype=dtype)

    def __call__(self, x: Float[Array, "seq_length d_inner"]):
        A = -jnp.exp(self.A_log)
        D = self.D

        delta_b_c = jax.vmap(self.input_proj)(x)

        split_indices = [
            self.dt_rank,
            self.dt_rank + self.d_state,
        ]
        delta, B, C = jnp.split(delta_b_c, split_indices, axis=-1)
        delta = jax.nn.softplus(jax.vmap(self.delta_proj)(delta))

        y = selective_scan(x, delta, A, B, C, D)
        return y


class MambaBlock(eqx.Module):
    in_proj: eqx.nn.Linear
    conv1d: eqx.nn.Conv1d
    ssm: SelectiveStateSpaceModel
    out_proj: eqx.nn.Linear

    def __init__(
        self,
        n_embd: int,
        d_inner: int,
        dt_rank: int,
        d_conv: int,
        use_in_projection_bias: bool = True,
        use_conv_bias: bool = True,
        use_out_proj_bias: bool = True,
        ssm_use_delta_proj_bias: bool = False,
        ssm_use_input_proj_bias: bool = False,
        *,
        key: PRNGKeyArray,
    ):
        (
            key,
            linear_key,
            conv1d_key,
            ssm_key,
            out_proj_key,
        ) = jax.random.split(key, 5)

        self.in_proj = eqx.nn.Linear(
            n_embd,
            d_inner * 2,
            use_bias=use_in_projection_bias,
            key=linear_key,
        )

        self.conv1d = eqx.nn.Conv1d(
            in_channels=d_inner,
            out_channels=d_inner,
            kernel_size=d_conv,
            use_bias=use_conv_bias,
            groups=d_inner,
            padding=d_conv - 1,
            key=conv1d_key,
        )
        self.ssm = SelectiveStateSpaceModel(
            d_inner=d_inner,
            dt_rank=dt_rank,
            d_state=d_inner,
            use_delta_proj_bias=ssm_use_delta_proj_bias,
            use_input_proj_bias=ssm_use_input_proj_bias,
            key=ssm_key,
        )
        self.out_proj = eqx.nn.Linear(
            d_inner,
            n_embd,
            use_bias=use_out_proj_bias,
            key=out_proj_key,
        )

    def __call__(self, x: Array):
        seq_len, d = x.shape
        x_and_res = jax.vmap(self.in_proj)(x)

        (x, res) = jnp.split(x_and_res, 2, axis=-1)
        x = jnp.transpose(x)
        x = self.conv1d(x)[:, :seq_len]
        x = jnp.transpose(x)
        x = jax.nn.silu(x)

        y = self.ssm(x)
        y = y * jax.nn.silu(res)

        output = jax.vmap(self.out_proj)(y)
        return output


class ResidualBlock(eqx.Module):
    mamba_block: MambaBlock
    rns_norm: eqx.nn.RMSNorm

    def __init__(
        self,
        n_embd: int,
        d_inner: int,
        dt_rank: int,
        d_conv: int,
        use_in_projection_bias: bool = True,
        use_conv_bias: bool = True,
        use_out_proj_bias: bool = True,
        ssm_use_delta_proj_bias: bool = False,
        ssm_use_input_proj_bias: bool = False,
        *,
        key: PRNGKeyArray,
    ):
        self.mamba_block = MambaBlock(
            n_embd=n_embd,
            d_inner=d_inner,
            dt_rank=dt_rank,
            d_conv=d_conv,
            use_in_projection_bias=use_in_projection_bias,
            use_conv_bias=use_conv_bias,
            use_out_proj_bias=use_out_proj_bias,
            ssm_use_delta_proj_bias=ssm_use_delta_proj_bias,
            ssm_use_input_proj_bias=ssm_use_input_proj_bias,
            key=key,
        )
        self.rns_norm = eqx.nn.RMSNorm(n_embd)

    def __call__(
        self, x: Float[Array, "seq_len n_embd"], *, key: Optional[PRNGKeyArray] = None
    ) -> Array:
        return self.mamba_block(jax.vmap(self.rns_norm)(x)) + x


class Mamba(eqx.Module):
    layers: eqx.nn.Sequential
    normalization: eqx.nn.RMSNorm
    shared_emb_lm_head: eqx.nn.Shared

    def __init__(
        self,
        n_embd: int,
        n_dims: int,
        n_layers: int,
        d_state: int = 16,
        expand: int = 2,
        dt_rank: int | str = "auto",
        d_conv: int = 4,
        pad_vocab_size_multiple: int = 8,
        conv_bias: bool = True,
        bias: bool = False,
        d_inner: int | None = None,
        use_in_projection_bias: bool = True,
        use_conv_bias: bool = True,
        use_out_proj_bias: bool = True,
        ssm_use_delta_proj_bias: bool = False,
        ssm_use_input_proj_bias: bool = False,
        *,
        key: PRNGKeyArray,
    ):
        d_inner = int(expand * n_embd)
        if dt_rank == "auto":
            dt_rank = math.ceil(n_embd / d_state)

        if n_dims % pad_vocab_size_multiple != 0:
            n_dims += pad_vocab_size_multiple - n_dims % pad_vocab_size_multiple
        key, *subkeys = jax.random.split(key, n_layers + 3)
        assert d_inner is not None and isinstance(d_inner, int)
        assert dt_rank is not None and isinstance(dt_rank, int)
        self.layers = eqx.nn.Sequential(
            [
                ResidualBlock(
                    n_embd=n_embd,
                    d_inner=d_inner,
                    dt_rank=dt_rank,
                    d_conv=d_conv,
                    use_in_projection_bias=use_in_projection_bias,
                    use_conv_bias=use_conv_bias,
                    use_out_proj_bias=use_out_proj_bias,
                    ssm_use_delta_proj_bias=ssm_use_delta_proj_bias,
                    ssm_use_input_proj_bias=ssm_use_input_proj_bias,
                    key=subkeys[i],
                )
                for i in range(n_layers)
            ],
        )
        self.normalization = eqx.nn.RMSNorm(n_embd)

        embedding = eqx.nn.Embedding(n_dims, n_embd, key=subkeys[n_layers])
        lm_head = eqx.nn.Linear(
            n_embd,
            n_dims,
            use_bias=False,
            key=subkeys[n_layers + 1],
        )
        where = lambda embed_and_lin: embed_and_lin[1].weight
        get = lambda embed_and_lin: embed_and_lin[0].weight
        self.shared_emb_lm_head = eqx.nn.Shared(
            (embedding, lm_head), where=where, get=get
        )

    def __call__(
        self,
        x: Int[Array, " seq_len"],
        *,
        state: eqx.nn.State | None = None,
        key: Optional[PRNGKeyArray] = None,
    ) -> Float[Array, "seq_len n_dims"]:
        embedding, linear = self.shared_emb_lm_head()
        x = eqx.filter_vmap(embedding)(x)
        x = self.layers(x)
        x = eqx.filter_vmap(self.normalization)(x)
        logits = eqx.filter_vmap(linear)(x)
        return logits



import functools

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from statedict2pytree import (
    convert,
    pytree_to_fields,
    state_dict_to_fields,
)
from torchvision.models.convnext import CNBlock as TorchCNBlock

from jaxonmodels.models.convnext import CNBlock


# Define the different parameter sets you want to test
@pytest.mark.parametrize(
    "batch_size, dim, layer_scale, stochastic_depth_prob, height, width",
    [
        # Original parameters
        (4, 128, 1e-6, 0.0, 56, 56),
        # Different dimension
        (4, 64, 1e-6, 0.0, 56, 56),
        # Different layer scale
        (4, 128, 1e-4, 0.0, 56, 56),
        # With stochastic depth
        (2, 128, 1e-6, 0.1, 56, 56),
        # Different input size
        (4, 128, 1e-6, 0.0, 28, 28),
        # Single batch item
        (1, 96, 1e-6, 0.0, 40, 40),
        # Combined changes
        (2, 64, 1e-5, 0.05, 32, 32),
    ],
)
def test_cnblock_equivalence_parametrized(  # Renamed function slightly
    batch_size,
    dim,
    layer_scale,
    stochastic_depth_prob,
    height,
    width,  # Added params as args
):
    # Keep norm_layer fixed as requested
    norm_layer_torch = None
    norm_layer_eqx = None

    key = jax.random.PRNGKey(42)
    # Use different keys for each test run potentially by splitting from a base key
    # Or reuse like this if deterministic randomness per parameter set is okay.
    key, eqx_init_key, data_key, vmap_key = jax.random.split(key, 4)

    dtype = jnp.float32
    np_dtype = np.float32
    torch_dtype = torch.float32

    # Use the parameters in the model instantiation
    torch_block = TorchCNBlock(
        dim, layer_scale, stochastic_depth_prob, norm_layer=norm_layer_torch
    )

    jax_block, state = eqx.nn.make_with_state(CNBlock)(
        dim,
        layer_scale,
        stochastic_depth_prob,
        norm_layer=norm_layer_eqx,  # Keep Equinox LayerNorm
        key=eqx_init_key,
        dtype=dtype,
    )

    weights_dict = torch_block.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    [print(t) for t in torchfields]
    jaxfields, state_indices = pytree_to_fields((jax_block, state))
    [print(j) for j in jaxfields]

    # Use the convert function
    jax_block, state = convert(
        weights_dict,
        (jax_block, state),
        jaxfields,
        state_indices,
        torchfields,
        dtype=dtype,
    )

    jax_block, state = eqx.nn.inference_mode((jax_block, state))

    # Use parameters for input shape
    np_input = np.random.rand(batch_size, dim, height, width).astype(np_dtype)
    torch_input = torch.from_numpy(np_input).to(torch_dtype)

    torch_block.eval()
    with torch.no_grad():
        torch_output = torch_block(torch_input)

    jax_input = jnp.array(np_input, dtype=dtype)
    # Ensure key is handled correctly if needed per-batch-item
    # If the model uses dropout/rng during inference based on the key,
    # you might need a different key handling strategy for vmap
    # For standard inference CNBlock, this partial should be fine.
    key, subkey = jax.random.split(key)
    jax_infer_func = functools.partial(
        jax_block, key=subkey
    )  # Usually key is for dropout during training
    jax_output = eqx.filter_vmap(jax_infer_func)(jax_input)

    assert jax_output.shape == torch_output.shape
    # You might need to adjust tolerance (atol/rtol) depending on params
    assert np.allclose(np.array(jax_output), torch_output.numpy(), atol=1e-5, rtol=1e-5)



from itertools import repeat

import jax
import jax.numpy as jnp
from beartype.typing import Any
from jaxtyping import Array, Float, PRNGKeyArray


def make_divisible(v: float, divisor: int, min_value: int | None = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def make_ntuple(x: Any, n: int) -> tuple[Any, ...]:
    if isinstance(x, collections.abc.Iterable):  # ty:ignore[unresolved-reference]
        return tuple(x)
    return tuple(repeat(x, n))


def default_floating_dtype():
    if jax.config.read("jax_enable_x64"):
        return jnp.float64
    else:
        return jnp.float32


def dtype_to_str(dtype) -> str:
    if hasattr(dtype, "__name__"):
        # For simple types like jnp.float32, float, etc.
        return dtype.__name__

    # For more complex types like JaxArray with specific dtype
    dtype_str = str(dtype)

    # Remove common patterns to clean up the string
    if "<class '" in dtype_str:
        # Extract the name from "<class 'jax.numpy.float32'>" -> "float32"
        dtype_str = dtype_str.split("'")[1].split(".")[-1]

    return dtype_str


def patch_merging_pad(x: Float[Array, "H W C"]) -> Array:
    H, W, C = x.shape
    h_pad = H % 2
    w_pad = W % 2
    x = jnp.pad(x, ((0, h_pad), (0, w_pad), (0, 0)))
    x0 = x[0::2, 0::2, :]
    x1 = x[1::2, 0::2, :]
    x2 = x[0::2, 1::2, :]
    x3 = x[1::2, 1::2, :]
    x = jnp.concatenate([x0, x1, x2, x3], axis=-1)
    return x


def default_init(
    key: PRNGKeyArray, shape: tuple[int, ...], dtype: Any, lim: float
) -> jax.Array:
    if jnp.issubdtype(dtype, jnp.complexfloating):
        real_dtype = jnp.finfo(dtype).dtype
        rkey, ikey = jax.random.split(key, 2)
        real = jax.random.uniform(rkey, shape, real_dtype, minval=-lim, maxval=lim)
        imag = jax.random.uniform(ikey, shape, real_dtype, minval=-lim, maxval=lim)
        return real.astype(dtype) + 1j * imag.astype(dtype)
    else:
        return jax.random.uniform(key, shape, dtype, minval=-lim, maxval=lim)



import functools
import math
import os
from abc import ABC
from functools import cache
from pathlib import Path

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from beartype.typing import Any, Literal, Self, Type, Union, cast
from jaxonlayers.functions.activation import swiglu
from jaxonlayers.functions.normalization import normalize
from jaxonlayers.layers.embedding import EmbeddingBag, EmbeddingWithPadding
from jaxtyping import Array, Float, Int, PRNGKeyArray
from statedict2pytree.converter import autoconvert

from jaxonmodels.functions import default_floating_dtype

ESM3_OPEN_SMALL = "esm3_sm_open_v1"
ESM3_OPEN_SMALL_ALIAS_1 = "esm3-open-2024-03"
ESM3_OPEN_SMALL_ALIAS_2 = "esm3-sm-open-v1"
ESM3_OPEN_SMALL_ALIAS_3 = "esm3-open"
ESM3_STRUCTURE_ENCODER_V0 = "esm3_structure_encoder_v0"
ESM3_STRUCTURE_DECODER_V0 = "esm3_structure_decoder_v0"
ESM3_FUNCTION_DECODER_V0 = "esm3_function_decoder_v0"
ESMC_600M = "esmc_600m"
ESMC_300M = "esmc_300m"


SEQUENCE_MASK_TOKEN = 32
SEQUENCE_BOS_TOKEN = 0
SEQUENCE_PAD_TOKEN = 1
SEQUENCE_EOS_TOKEN = 2
SEQUENCE_CHAINBREAK_TOKEN = 31
STRUCTURE_MASK_TOKEN = 4096
STRUCTURE_PAD_TOKEN = 4099
STRUCTURE_BOS_TOKEN = 4098
STRUCTURE_EOS_TOKEN = 4097
STRUCTURE_CHAINBREAK_TOKEN = 4100
SS8_PAD_TOKEN = 0
SASA_PAD_TOKEN = 0
RESIDUE_PAD_TOKEN = 0
INTERPRO_PAD_TOKEN = 0
SEQUENCE_VOCAB = [
    "<cls>",
    "<pad>",
    "<eos>",
    "<unk>",
    "L",
    "A",
    "G",
    "V",
    "S",
    "E",
    "R",
    "T",
    "I",
    "D",
    "P",
    "K",
    "Q",
    "N",
    "F",
    "Y",
    "M",
    "H",
    "W",
    "C",
    "X",
    "B",
    "U",
    "Z",
    "O",
    ".",
    "-",
    "|",
    "<mask>",
]
SEQUENCE_BOS_TOKEN = 0
SEQUENCE_EOS_TOKEN = 2


@staticmethod
@cache
def data_root(model: str):
    from huggingface_hub import snapshot_download

    if "INFRA_PROVIDER" in os.environ:
        return Path("")
    # Try to download from hugginface if it doesn't exist
    if model.startswith("esm3"):
        path = Path(snapshot_download(repo_id="EvolutionaryScale/esm3-sm-open-v1"))
    elif model.startswith("esmc-300"):
        path = Path(snapshot_download(repo_id="EvolutionaryScale/esmc-300m-2024-12"))
    elif model.startswith("esmc-600"):
        path = Path(snapshot_download(repo_id="EvolutionaryScale/esmc-600m-2024-12"))
    else:
        raise ValueError(f"{model=} is an invalid model name.")
    return path


def _graham_schmidt(x_axis: Array, xy_plane: Array, eps: float = 1e-12):
    e1 = xy_plane
    denom = jnp.sqrt((x_axis**2).sum(axis=-1, keepdims=True) + eps)
    x_axis = x_axis / denom
    dot = (x_axis * e1).sum(axis=-1, keepdims=True)
    e1 = e1 - x_axis * dot
    denom = jnp.sqrt((e1**2).sum(axis=-1, keepdims=True) + eps)
    e1 = e1 / denom
    e2 = jnp.cross(x_axis, e1, axis=-1)
    rots = jnp.stack([x_axis, e1, e2], axis=-1)
    return rots


def _sqrt_subgradient(x: Array) -> Array:
    return jnp.where(x > 0, jnp.sqrt(x), 0.0)


def _quat_invert(q: Array):
    return q * jnp.array([1, -1, -1, -1])


def _quat_rotation(q: Float[Array, "w x y z"], p: Float[Array, "x y z"]) -> Array:
    """
    Rotates p by quaternion q.

    Args:
        q: Quaternions as tensor of shape (..., 4), real part first.
        p: Points as tensor of shape (..., 3)

    Returns:
        The rotated version of p, of shape (..., 3)
    """
    aw, ax, ay, az = q[..., 0], q[..., 1], q[..., 2], q[..., 3]
    bx, by, bz = p[..., 0], p[..., 1], p[..., 2]
    # fmt: off
    ow =         - ax * bx - ay * by - az * bz
    ox = aw * bx           + ay * bz - az * by
    oy = aw * by - ax * bz           + az * bx
    oz = aw * bz + ax * by - ay * bx
    # fmt: on
    q_mul_pts = jnp.stack((ow, ox, oy, oz), -1)
    return _quat_mult(q_mul_pts, _quat_invert(q))[..., 1:]


def _quat_mult(
    a: Float[Array, "w x y z"], b: Float[Array, "w x y z"]
) -> Float[Array, "w x y z"]:
    """
    Multiply two quaternions.
    Usual torch rules for broadcasting apply.

    Args:
        a: Quaternions as tensor of shape (..., 4), real part first.
        b: Quaternions as tensor of shape (..., 4), real part first.

    Returns:
        The product of a and b, a tensor of quaternions shape (..., 4).
    """
    aw, ax, ay, az = a[..., 0], a[..., 1], a[..., 2], a[..., 3]
    bw, bx, by, bz = b[..., 0], b[..., 1], b[..., 2], b[..., 3]
    ow = aw * bw - ax * bx - ay * by - az * bz
    ox = aw * bx + ax * bw + ay * bz - az * by
    oy = aw * by - ax * bz + ay * bw + az * bx
    oz = aw * bz + ax * by - ay * bx + az * bw
    return jnp.stack((ow, ox, oy, oz), -1)


class Rotation(ABC):
    @classmethod
    def identity(cls, shape: tuple[int, ...], **tensor_kwargs) -> Self:
        raise NotImplementedError

    @classmethod
    def random(cls, shape: tuple[int, ...], key: PRNGKeyArray, **tensor_kwargs) -> Self:
        raise NotImplementedError

    def __getitem__(self, idx: Any) -> Self:
        raise NotImplementedError

    @property
    def tensor(self) -> Array:
        raise NotImplementedError

    @property
    def shape(self) -> tuple:
        raise NotImplementedError

    def as_matrix(self) -> "RotationMatrix":
        raise NotImplementedError

    def as_quat(self, normalize: bool = False) -> "RotationQuat":
        raise NotImplementedError

    def compose(self, other: "Rotation") -> "Rotation":
        raise NotImplementedError

    def convert_compose(self, other: Self) -> Self:
        raise NotImplementedError

    def apply(self, p: Array) -> Array:
        raise NotImplementedError

    def invert(self) -> Self:
        raise NotImplementedError

    @property
    def dtype(self) -> jnp.dtype:
        return self.tensor.dtype

    @property
    def device(self) -> jax.Device:
        return self.tensor.device

    def tensor_apply(self, func) -> Self:
        # Applys a function to the underlying tensor
        return type(self)(
            jnp.stack(
                [func(self.tensor[..., i]) for i in range(self.tensor.shape[-1])],
                axis=-1,
            )  # ty: ignore too-many-positional-arguments
        )


class RotationQuat(Rotation, eqx.Module):
    _quats: Array
    _normalized: bool = eqx.field(static=True)

    def __init__(self, quats: Array, normalized=False):
        assert quats.shape[-1] == 4
        self._normalized = normalized
        # Force float32 as well
        if normalized:
            self._quats = normalize(quats.astype(jnp.float32), axis=-1)
            self._quats = jnp.where(
                self._quats[..., :1] >= 0, self._quats, -self._quats
            )
        else:
            self._quats = quats.astype(jnp.float32)

    @classmethod
    def identity(cls, shape, **tensor_kwargs):
        q = jnp.ones((*shape, 4), **tensor_kwargs)
        mult = jnp.array([1, 0, 0, 0])
        return RotationQuat(q * mult)

    @classmethod
    def random(cls, shape, key: PRNGKeyArray, **tensor_kwargs):
        quat = jax.random.normal(key=key, shape=(*shape, 4), **tensor_kwargs)
        return RotationQuat(quat, normalized=True)

    def __getitem__(self, idx: Any) -> "RotationQuat":
        if isinstance(idx, (int, slice)) or idx is None:
            indices = (idx,)
        else:
            indices = tuple(idx)
        return RotationQuat(self._quats[indices + (slice(None),)])

    @property
    def shape(self) -> tuple:
        return self._quats.shape[:-1]

    def compose(self, other: Rotation) -> Rotation:
        assert isinstance(other, RotationQuat)
        return RotationQuat(_quat_mult(self._quats, other._quats))

    def convert_compose(self, other: Rotation):
        return self.compose(other.as_quat())

    def as_matrix(self) -> "RotationMatrix":
        q = self.normalized().tensor
        r, i, j, k = q[..., 0], q[..., 1], q[..., 2], q[..., 3]
        two_s = 2.0 / jnp.linalg.norm(q, axis=-1)

        o = jnp.stack(
            (
                1 - two_s * (j * j + k * k),
                two_s * (i * j - k * r),
                two_s * (i * k + j * r),
                two_s * (i * j + k * r),
                1 - two_s * (i * i + k * k),
                two_s * (j * k - i * r),
                two_s * (i * k - j * r),
                two_s * (j * k + i * r),
                1 - two_s * (i * i + j * j),
            ),
            -1,
        )
        return RotationMatrix(o.reshape(q.shape[:-1] + (3, 3)))

    def as_quat(self, normalize: bool = False) -> "RotationQuat":
        return self

    def apply(self, p: Array) -> Array:
        return _quat_rotation(self.normalized()._quats, p)

    def invert(self) -> "RotationQuat":
        return RotationQuat(_quat_invert(self._quats))

    @property
    def tensor(self) -> Array:
        return self._quats

    def normalized(self) -> "RotationQuat":
        return self if self._normalized else RotationQuat(self._quats, normalized=True)


class RotationMatrix(Rotation, eqx.Module):
    _rots: Array

    def __init__(self, rots: Array):
        if rots.shape[-1] == 9:
            rots = rots.reshape(*rots.shape[:-1], 3, 3)
        assert rots.shape[-1] == 3
        assert rots.shape[-2] == 3
        # Force full precision
        rots = rots.astype(jnp.float32)
        self._rots = rots

    @classmethod
    def identity(cls, shape, **tensor_kwargs):
        rots = jnp.eye(3, **tensor_kwargs)
        rots = rots.reshape(*[1 for _ in range(len(shape))], 3, 3)
        rots = jnp.broadcast_to(rots, (*shape, rots.shape[-2], rots.shape[-1]))
        return cls(rots)

    @classmethod
    def random(cls, shape, key: PRNGKeyArray, **tensor_kwargs):
        return RotationQuat.random(shape, **tensor_kwargs).as_matrix()

    def __getitem__(self, idx: Any) -> "RotationMatrix":
        indices = (idx,) if isinstance(idx, int) or idx is None else tuple(idx)
        return RotationMatrix(self._rots[indices + (slice(None), slice(None))])

    @property
    def shape(self) -> tuple:
        return self._rots.shape[:-2]

    def as_matrix(self) -> "RotationMatrix":
        return self

    def as_quat(self, normalize: bool = False) -> RotationQuat:
        flat = self._rots.reshape(*self._rots.shape[:-2], 9)
        m00, m01, m02, m10, m11, m12, m20, m21, m22 = [flat[..., i] for i in range(9)]
        q_abs = _sqrt_subgradient(
            jnp.stack(
                [
                    1.0 + m00 + m11 + m22,
                    1.0 + m00 - m11 - m22,
                    1.0 - m00 + m11 - m22,
                    1.0 - m00 - m11 + m22,
                ],
                axis=-1,
            )
        )
        # we produce the desired quaternion multiplied by each of r, i, j, k
        quat_by_rijk = jnp.stack(
            [
                x
                for lst in [
                    [q_abs[..., 0] ** 2, m21 - m12, m02 - m20, m10 - m01],
                    [m21 - m12, q_abs[..., 1] ** 2, m10 + m01, m02 + m20],
                    [m02 - m20, m10 + m01, q_abs[..., 2] ** 2, m12 + m21],
                    [m10 - m01, m20 + m02, m21 + m12, q_abs[..., 3] ** 2],
                ]
                for x in lst
            ],
            axis=-1,
        )
        quat_by_rijk = quat_by_rijk.reshape(*quat_by_rijk.shape[:-1], 4, 4)

        # We floor here at 0.1 but the exact level is not important; if q_abs is small,
        # the candidate won't be picked.
        flr = jnp.array(0.1, dtype=q_abs.dtype)
        quat_candidates = quat_by_rijk / (2.0 * jnp.maximum(q_abs[..., None], flr))

        one_hot = jax.nn.one_hot(q_abs.argmax(axis=-1), num_classes=q_abs.shape[-1])
        quat = (quat_candidates * one_hot[..., :, None]).sum(axis=-2)

        return RotationQuat(quat)

    def compose(self, other: Rotation) -> Rotation:
        assert isinstance(other, RotationMatrix)
        return RotationMatrix(self._rots @ other._rots)

    def convert_compose(self, other: Rotation):
        return self.compose(other.as_matrix())

    def apply(self, p: Array) -> Array:
        if self._rots.shape[-3] == 1:
            # This is a slight speedup over einsum for batched rotations
            return p @ jnp.swapaxes(self._rots, -1, -2).squeeze(-3)
        else:
            # einsum way faster than bmm!
            return jnp.einsum("...ij,...j", self._rots, p)

    def invert(self) -> "RotationMatrix":
        return RotationMatrix(jnp.swapaxes(self._rots, -1, -2))

    @property
    def tensor(self) -> Array:
        return self._rots.reshape(*self._rots.shape[:-2], -1)

    def to_3x3(self) -> Array:
        return self._rots

    @staticmethod
    def from_graham_schmidt(
        x_axis: Array, xy_plane: Array, eps: float = 1e-12
    ) -> "RotationMatrix":
        return RotationMatrix(_graham_schmidt(x_axis, xy_plane, eps))


class Affine3D(eqx.Module):
    trans: Array
    rot: Rotation

    def __post_init__(self):
        assert self.trans.shape[:-1] == self.rot.shape

    @staticmethod
    def identity(
        shape_or_affine: Union[tuple[int, ...], "Affine3D"],
        rotation_type: Type[Rotation] = RotationMatrix,
        **tensor_kwargs,
    ):
        if isinstance(shape_or_affine, Affine3D):
            shape = shape_or_affine.shape
            rotation_type = type(shape_or_affine.rot)
            dtype = shape_or_affine.dtype
        else:
            shape = shape_or_affine
            dtype = tensor_kwargs.get("dtype", None)

        return Affine3D(
            jnp.zeros((*shape, 3), dtype=dtype),
            rotation_type.identity(shape, dtype=dtype),
        )

    @staticmethod
    def random(
        key: PRNGKeyArray,
        shape: tuple[int, ...],
        std: float = 1,
        rotation_type: Type[Rotation] = RotationMatrix,
        **tensor_kwargs,
    ) -> "Affine3D":
        return Affine3D(
            trans=jax.random.normal(key=key, shape=(*shape, 3), **tensor_kwargs) * std,
            rot=rotation_type.random(shape, **tensor_kwargs),
        )

    def __getitem__(self, idx: Any) -> "Affine3D":
        indices = (idx,) if isinstance(idx, int) or idx is None else tuple(idx)
        return Affine3D(trans=self.trans[indices + (slice(None),)], rot=self.rot[idx])

    @property
    def shape(self) -> tuple:
        return self.trans.shape[:-1]

    @property
    def dtype(self) -> jnp.dtype:
        return self.trans.dtype

    @property
    def device(self) -> jax.Device:
        return self.trans.device

    def tensor_apply(self, func) -> "Affine3D":
        # Applys a function to the underlying tensor
        return self.from_tensor(
            jnp.stack(
                [func(self.tensor[..., i]) for i in range(self.tensor.shape[-1])],
                axis=-1,
            )
        )

    def as_matrix(self):
        return Affine3D(trans=self.trans, rot=self.rot.as_matrix())

    def as_quat(self, normalize: bool = False):
        return Affine3D(trans=self.trans, rot=self.rot.as_quat(normalize))

    def compose(self, other: "Affine3D", autoconvert: bool = False):
        rot = self.rot
        new_rot = (rot.convert_compose if autoconvert else rot.compose)(other.rot)
        new_trans = rot.apply(other.trans) + self.trans
        return Affine3D(trans=new_trans, rot=new_rot)

    def compose_rotation(self, other: Rotation, autoconvert: bool = False):
        return Affine3D(
            trans=self.trans,
            rot=(self.rot.convert_compose if autoconvert else self.rot.compose)(other),
        )

    def scale(self, v: Array | float):
        return Affine3D(self.trans * v, self.rot)

    def mask(self, mask: Array, with_zero=False):
        # Returns a transform where True positions in mask is identity
        if with_zero:
            tensor = self.tensor
            return Affine3D.from_tensor(
                jnp.where(mask[..., None], jnp.zeros_like(tensor), tensor)
            )
        else:
            identity = self.identity(
                self.shape,
                rotation_type=type(self.rot),
                dtype=self.dtype,
            ).tensor
            return Affine3D.from_tensor(
                jnp.where(mask[..., None], identity, self.tensor)
            )

    def apply(self, p: Array) -> Array:
        return self.rot.apply(p) + self.trans

    def invert(self):
        inv_rot = self.rot.invert()
        return Affine3D(trans=-inv_rot.apply(self.trans), rot=inv_rot)

    @property
    def tensor(self) -> Array:
        return jnp.concat([self.rot.tensor, self.trans], axis=-1)

    @staticmethod
    def from_tensor(t: Array) -> "Affine3D":
        match t.shape[-1]:
            case 4:
                # Assume tensor 4x4 for backward compat with alphafold
                trans = t[..., :3, 3]
                rot = RotationMatrix(t[..., :3, :3])
            case 6:
                trans = t[..., -3:]
                x = t[..., :3]
                padded = jnp.concatenate([jnp.ones((*x.shape[:-1], 1)), x], axis=-1)
                rot = RotationQuat(padded)
            case 7:
                trans = t[..., -3:]
                rot = RotationQuat(t[..., :4])
            case 12:
                trans = t[..., -3:]
                rot = RotationMatrix(t[..., :-3].reshape(*t.shape[:-1], 3, 3))
            case _:
                raise RuntimeError(
                    "Cannot detect rotation fromat "
                    f"from {t.shape[-1] - 3}-d flat vector"
                )
        return Affine3D(trans, rot)

    @staticmethod
    def from_tensor_pair(t: Array, r: Array) -> "Affine3D":
        return Affine3D(t, RotationMatrix(r))

    @staticmethod
    def from_graham_schmidt(
        neg_x_axis: Array,
        origin: Array,
        xy_plane: Array,
        eps: float = 1e-10,
    ):
        # The arguments of this function is for parity with AlphaFold
        x_axis = origin - neg_x_axis
        xy_plane = xy_plane - origin
        return Affine3D(
            trans=origin, rot=RotationMatrix.from_graham_schmidt(x_axis, xy_plane, eps)
        )

    @staticmethod
    def cat(affines: list["Affine3D"], dim: int = 0):
        if dim < 0:
            dim = len(affines[0].shape) + dim
        return Affine3D.from_tensor(jnp.concat([x.tensor for x in affines], axis=dim))


def swiglu_correction_fn(expansion_ratio: float, d_model: int) -> int:
    # set hidden dimesion to nearest multiple of 256 after expansion ratio
    return int(((expansion_ratio * d_model) + 255) // 256 * 256)


def swiglu_ln_ffn(
    d_model: int,
    expansion_ratio: float,
    bias: bool,
    *,
    key: PRNGKeyArray,
    dtype: Any | None,
):
    key1, key2 = jax.random.split(key)
    return [
        eqx.nn.LayerNorm(d_model),
        eqx.nn.Linear(
            d_model,
            swiglu_correction_fn(expansion_ratio, d_model) * 2,
            use_bias=bias,
            key=key1,
            dtype=dtype,
        ),
        eqx.nn.Lambda(fn=swiglu),
        eqx.nn.Linear(
            swiglu_correction_fn(expansion_ratio, d_model),
            d_model,
            use_bias=bias,
            key=key2,
            dtype=dtype,
        ),
    ]


def gelu_ln_ffn(
    d_model: int,
    expansion_ratio: float,
    bias: bool,
    *,
    key: PRNGKeyArray,
    dtype: Any | None,
):
    hidden_dim = int(expansion_ratio * d_model)
    key1, key2 = jax.random.split(key)
    return [
        eqx.nn.LayerNorm(d_model),
        eqx.nn.Linear(d_model, hidden_dim, use_bias=bias, key=key1, dtype=dtype),
        eqx.nn.Lambda(fn=jax.nn.gelu),
        eqx.nn.Linear(hidden_dim, d_model, use_bias=bias, key=key2, dtype=dtype),
    ]


class GeometricReasoningOriginalImpl(eqx.Module):
    s_norm: eqx.nn.LayerNorm
    proj: eqx.nn.Linear
    distance_scale_per_head: Array
    rotation_scale_per_head: Array
    out_proj: eqx.nn.Linear

    v_heads: int = eqx.field(static=True)
    num_vector_messages: int = eqx.field(static=True)
    mask_and_zero_frameless: bool = eqx.field(static=True)

    inference: bool

    def __init__(
        self,
        c_s: int,
        v_heads: int,
        num_vector_messages: int = 1,
        mask_and_zero_frameless: bool = True,
        bias: bool = False,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.inference = inference

        self.v_heads = v_heads
        self.num_vector_messages = num_vector_messages
        self.mask_and_zero_frameless = mask_and_zero_frameless

        self.s_norm = eqx.nn.LayerNorm(c_s, use_bias=bias)
        dim_proj = 4 * self.v_heads * 3 + self.v_heads * 3 * self.num_vector_messages
        # 2 x (q, k) * number of heads * (x, y, z) +
        #   number of heads * number of vector messages * (x, y, z)

        key, proj_key, out_proj_key = jax.random.split(key, 3)
        self.proj = eqx.nn.Linear(
            c_s, dim_proj, use_bias=bias, key=proj_key, dtype=dtype
        )
        channels_out = self.v_heads * 3 * self.num_vector_messages
        self.out_proj = eqx.nn.Linear(
            channels_out, c_s, key=out_proj_key, use_bias=bias, dtype=dtype
        )
        self.distance_scale_per_head = jnp.zeros((self.v_heads))
        self.rotation_scale_per_head = jnp.zeros((self.v_heads))

    def __call__(
        self,
        s: Array,
        affine: Affine3D,
        affine_mask: Array,
        sequence_id: Array | None,
        chain_id: Array,
    ):
        seq_len = s.shape[0]

        if sequence_id is None:
            if jax.config.read("jax_enable_x64"):
                dtype = jnp.int64
            else:
                dtype = jnp.int32
            sequence_id = jnp.zeros(seq_len, dtype=dtype)

        attn_bias = sequence_id[:, None] == sequence_id[None, :]
        attn_bias = attn_bias[None, :, :].astype(s.dtype)
        attn_bias = jnp.where(
            affine_mask[None, None, :],
            attn_bias,
            jnp.finfo(attn_bias.dtype).min,
        )

        chain_id_mask = chain_id[:, None] != chain_id[None, :]
        attn_bias = jnp.where(
            chain_id_mask[None, :, :],
            jnp.finfo(s.dtype).min,
            attn_bias,
        )
        ns = eqx.filter_vmap(self.s_norm)(s)
        proj_out = eqx.filter_vmap(self.proj)(ns)

        split_idx = self.v_heads * 2 * 3 + self.v_heads * 3 * self.num_vector_messages
        vec_rot = proj_out[..., :split_idx]
        vec_dist = proj_out[..., split_idx:]

        vec_rot = vec_rot.reshape(seq_len, -1, 3)
        rotated = affine.rot[..., None].apply(vec_rot)

        qr_end = self.v_heads
        kr_end = self.v_heads * 2
        query_rot = rotated[:, :qr_end, :]
        key_rot = rotated[:, qr_end:kr_end, :]
        value = rotated[:, kr_end:, :]

        vec_dist = vec_dist.reshape(seq_len, -1, 3)
        transformed = affine[..., None].apply(vec_dist)
        query_dist, key_dist = jnp.split(transformed, 2, axis=-2)

        query_dist = jnp.transpose(query_dist, (1, 0, 2))[:, :, None, :]
        key_dist = jnp.transpose(key_dist, (1, 0, 2))[:, None, :, :]
        query_rot = jnp.transpose(query_rot, (1, 0, 2))
        key_rot = jnp.transpose(key_rot, (1, 2, 0))

        value = value.reshape(seq_len, self.v_heads, self.num_vector_messages, 3)
        value = jnp.transpose(value, (1, 0, 2, 3))
        value = value.reshape(self.v_heads, seq_len, self.num_vector_messages * 3)

        distance_term = jnp.linalg.norm(query_dist - key_dist, axis=-1) / math.sqrt(3)
        rotation_term = query_rot @ key_rot / math.sqrt(3)

        distance_term_weight = jax.nn.softplus(self.distance_scale_per_head)[
            :, None, None
        ]
        rotation_term_weight = jax.nn.softplus(self.rotation_scale_per_head)[
            :, None, None
        ]

        attn_weight = (
            rotation_term * rotation_term_weight - distance_term * distance_term_weight
        )

        if attn_bias is not None:
            s_q = attn_weight.shape[1]
            s_k = attn_weight.shape[2]
            _s_q = max(0, attn_bias.shape[1] - s_q)
            _s_k = max(0, attn_bias.shape[2] - s_k)
            attn_bias = attn_bias[:, _s_q:, _s_k:]
            attn_weight = attn_weight + attn_bias

        attn_weight = jnp.nan_to_num(jax.nn.softmax(attn_weight, axis=-1), nan=0.0)

        attn_out = attn_weight @ value

        attn_out = attn_out.reshape(self.v_heads, seq_len, self.num_vector_messages, 3)
        attn_out = jnp.transpose(attn_out, (1, 0, 2, 3))
        attn_out = attn_out.reshape(seq_len, self.v_heads * self.num_vector_messages, 3)

        attn_out = affine.rot[..., None].invert().apply(attn_out)

        attn_out = attn_out.reshape(
            seq_len, self.v_heads * self.num_vector_messages * 3
        )

        if self.mask_and_zero_frameless:
            attn_out = jnp.where(affine_mask[:, None], attn_out, 0.0)

        s = eqx.filter_vmap(self.out_proj)(attn_out)
        return s


class ESMMultiHeadAttention(eqx.Module):
    d_model: int = eqx.field(static=True)
    n_heads: int = eqx.field(static=True)
    d_head: int = eqx.field(static=True)
    qk_layernorm: bool = eqx.field(static=True)

    layernorm_qkv: list

    q_ln: eqx.nn.LayerNorm | None
    k_ln: eqx.nn.LayerNorm | None
    rotary: eqx.nn.RotaryPositionalEmbedding
    out_proj: eqx.nn.Linear

    inference: bool

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        use_bias: bool = False,
        qk_layernorm: bool = True,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.inference = inference

        self.d_model = d_model
        self.n_heads = n_heads
        self.d_head = d_model // n_heads
        self.qk_layernorm = qk_layernorm

        key, *subkeys = jax.random.split(key, 3)

        self.layernorm_qkv = [
            eqx.nn.LayerNorm(d_model),
            eqx.nn.Linear(
                d_model, d_model * 3, use_bias=use_bias, key=subkeys[0], dtype=dtype
            ),
        ]

        if qk_layernorm:
            self.q_ln = eqx.nn.LayerNorm(d_model, use_bias=use_bias)
            self.k_ln = eqx.nn.LayerNorm(d_model, use_bias=use_bias)
        else:
            self.q_ln = None
            self.k_ln = None

        self.rotary = eqx.nn.RotaryPositionalEmbedding(self.d_head)
        self.out_proj = eqx.nn.Linear(
            d_model, d_model, use_bias=use_bias, key=subkeys[1], dtype=dtype
        )

    def __call__(
        self,
        x: Float[Array, "seq_len d_model"],
        seq_id: Int[Array, "seq_len"] | None = None,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Float[Array, "seq_len d_model"]:
        seq_len = x.shape[0]
        dtype = x.dtype

        for l in self.layernorm_qkv:
            x = eqx.filter_vmap(l)(x)
        qkv = x
        q, k, v = jnp.split(qkv, 3, axis=-1)

        if self.q_ln is not None:
            q = eqx.filter_vmap(self.q_ln)(q).astype(dtype)
        if self.k_ln is not None:
            k = eqx.filter_vmap(self.k_ln)(k).astype(dtype)

        q = q.reshape(seq_len, self.n_heads, self.d_head)
        k = k.reshape(seq_len, self.n_heads, self.d_head)
        v = v.reshape(seq_len, self.n_heads, self.d_head)

        q = eqx.filter_vmap(self.rotary, in_axes=1, out_axes=1)(q)
        k = eqx.filter_vmap(self.rotary, in_axes=1, out_axes=1)(k)

        mask = None
        if seq_id is not None:
            mask = seq_id[:, None] == seq_id[None, :]
            mask = jnp.expand_dims(mask, axis=0)
        context = jax.nn.dot_product_attention(q, k, v, mask=mask)
        context = context.reshape(seq_len, self.d_model)

        output = eqx.filter_vmap(self.out_proj)(context)

        return output


def RegressionHead(
    d_model: int,
    output_dim: int,
    hidden_dim: int | None = None,
    *,
    key: PRNGKeyArray,
    dtype: Any | None = None,
) -> list:
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None
    hidden_dim = hidden_dim if hidden_dim is not None else d_model
    key, subkey = jax.random.split(key)
    return [
        eqx.nn.Linear(d_model, hidden_dim, key=key),
        eqx.nn.Lambda(fn=jax.nn.gelu),
        eqx.nn.LayerNorm(hidden_dim),
        eqx.nn.Linear(hidden_dim, output_dim, key=subkey),
    ]


class UnifiedTransformerBlock(eqx.Module):
    """
    A unified transformer block that can optionally incorporate geometric attention.

    This class defines a transformer block that can be configured to use
    geometric attention alongside the standard multi-head attention mechanism.
    It is designed to be a flexible component of transformer-based models, allowing for
    the integration of geometric reasoning.

    Parameters
    ----------
    d_model : int
        The dimensionality of the input and output features of the transformer block.
    n_heads : int
        The number of attention heads in the multi-head attention mechanism.
    n_layers : int
        The number of layers in the transformer block.
    use_geom_attn : bool, optional
        Whether to use geometric attention in addition to
        the standard multi-head attention. Defaults to False.
    v_heads : int, optional
        The number of heads to use for the geometric attention mechanism, if enabled.
        Must be specified if `use_geom_attn` is True.
    """

    attn: ESMMultiHeadAttention | None
    geom_attn: GeometricReasoningOriginalImpl | None
    ffn: list

    scaling_factor: float = eqx.field(static=True)

    use_plain_attn: bool = eqx.field(static=True)
    use_geom_attn: bool = eqx.field(static=True)

    inference: bool

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        use_geom_attn: bool = False,
        use_plain_attn: bool = True,
        v_heads: int | None = None,
        bias: bool = False,
        expansion_ratio: float = 4.0,
        residue_scaling_factor: float = 1,
        mask_and_zero_frameless: bool = False,
        qk_layernorm: bool = True,
        ffn_type: Literal["swiglu", "gelu"] = "swiglu",  # swiglu | gelu
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.inference = inference
        self.use_plain_attn = use_plain_attn
        self.use_geom_attn = use_geom_attn
        key, mha_key = jax.random.split(key)
        if self.use_plain_attn:
            self.attn = ESMMultiHeadAttention(
                d_model,
                n_heads,
                bias,
                qk_layernorm=qk_layernorm,
                key=mha_key,
                inference=inference,
                dtype=dtype,
            )
        else:
            self.attn = None

        if self.use_geom_attn:
            if v_heads is None:
                raise ValueError("v_heads must be specified when use_geom_attn is True")
            key, geom_key = jax.random.split(key)
            self.geom_attn = GeometricReasoningOriginalImpl(
                c_s=d_model,
                v_heads=v_heads,
                bias=bias,
                mask_and_zero_frameless=mask_and_zero_frameless,
                key=geom_key,
                dtype=dtype,
                inference=inference,
            )
        else:
            self.geom_attn = None
        key, ffn_key = jax.random.split(key)
        if ffn_type == "swiglu":
            self.ffn = swiglu_ln_ffn(
                d_model,
                expansion_ratio,
                bias,
                key=ffn_key,
                dtype=dtype,
            )
        elif ffn_type == "gelu":
            self.ffn = gelu_ln_ffn(
                d_model,
                expansion_ratio,
                bias,
                key=ffn_key,
                dtype=dtype,
            )
        else:
            raise ValueError(f"Unknown ffn_type: {ffn_type}")

        self.scaling_factor = residue_scaling_factor

    def __call__(
        self,
        x: Array,
        sequence_id: Array | None,
        frames: Affine3D | None,
        frames_mask: Array | None,
        chain_id: Array | None,
    ) -> Array:
        if self.use_plain_attn:
            assert self.attn is not None
            r1 = self.attn(x, sequence_id)
            x = x + r1 / self.scaling_factor

        if self.use_geom_attn:
            assert frames is not None
            assert frames_mask is not None
            assert self.geom_attn is not None
            assert chain_id is not None
            r2 = self.geom_attn(x, frames, frames_mask, sequence_id, chain_id)
            x = x + r2 / self.scaling_factor

        ffn_x = x
        for l in self.ffn:
            ffn_x = eqx.filter_vmap(l)(ffn_x)
        r3 = ffn_x / self.scaling_factor
        x = x + r3

        return x


class TransformerStack(eqx.Module):
    """
    A stack of transformer blocks used in the ESM-3 model.
    Each block is a UnifiedTransformerBlock,
    which can either be geometric attention or standard multi-head attention.

    Args:
        d_model (int): The dimensionality of the input and output feature vectors.
        n_heads (int): The number of attention heads.
        v_heads (int): The number of voting heads.
        n_layers (int): The number of transformer blocks in the stack.
        n_layers_geom (int, optional): The number of transformer blocks that use
            geometric attention.
        scale_residue (bool, optional): Whether to scale the residue connections
            in each transformer block.
        mask_and_zero_frameless (bool, optional): Whether to mask and zero frameless
            positions in the input.

        Only applies in the geometric attention blocks, which is
        conditioned on the structure
    """

    blocks: list[UnifiedTransformerBlock]
    norm: eqx.nn.LayerNorm

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        v_heads: int | None,
        n_layers: int,
        n_layers_geom: int = 1,
        scale_residue: bool = True,
        mask_and_zero_frameless: bool = False,
        bias: bool = False,
        qk_layernorm: bool = True,
        ffn_type: Literal["swiglu", "gelu"] = "swiglu",  # swiglu | gelu
        expansion_ratio: float = 8 / 3,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        keys = jax.random.split(key, n_layers)
        self.blocks = [
            UnifiedTransformerBlock(
                d_model,
                n_heads,
                v_heads=v_heads,
                use_geom_attn=i < n_layers_geom,
                residue_scaling_factor=(
                    math.sqrt(n_layers / 36) if scale_residue else 1.0
                ),
                expansion_ratio=expansion_ratio,
                mask_and_zero_frameless=mask_and_zero_frameless,
                bias=bias,
                qk_layernorm=qk_layernorm,
                ffn_type=ffn_type,
                key=keys[i],
                dtype=dtype,
                inference=inference,
            )
            for i in range(n_layers)
        ]
        self.norm = eqx.nn.LayerNorm(d_model, use_bias=False, dtype=dtype)

    def __call__(
        self,
        x: Array,
        sequence_id: Array | None = None,
        frames: Affine3D | None = None,
        frames_mask: Array | None = None,
        chain_id: Array | None = None,
    ) -> tuple[Array, Array, list[Array]]:
        hiddens = []
        for block in self.blocks:
            x = block(x, sequence_id, frames, frames_mask, chain_id)
            hiddens.append(x)
        embedding = x
        x = eqx.filter_vmap(self.norm)(x)
        return x, embedding, hiddens


class ESMC(eqx.Module):
    """
    ESMC model implementation.

    Args:
        d_model (int): The dimensionality of the input and output feature vectors.
        n_heads (int): The number of attention heads in the transformer layers.
        n_layers (int): The number of transformer layers.
    """

    embed: eqx.nn.Embedding
    transformer: TransformerStack
    sequence_head: list

    inference: bool

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        n_layers: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if not dtype:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.inference = inference
        key, embed_key, transformer_stack_key, regression_head_key = jax.random.split(
            key, 4
        )
        self.embed = eqx.nn.Embedding(64, d_model, key=embed_key)

        self.transformer = TransformerStack(
            d_model,
            n_heads,
            None,
            n_layers,
            n_layers_geom=0,
            key=transformer_stack_key,
            dtype=dtype,
            inference=inference,
        )

        self.sequence_head = RegressionHead(d_model, 64, key=regression_head_key)

    def __call__(
        self,
        sequence_tokens: Array,
        sequence_id: Array | None = None,
    ) -> tuple[Array, Array, Array]:
        if sequence_id is None:
            # 1 is the PAD_TOKEN_ID
            sequence_id = sequence_tokens != 1

        x = eqx.filter_vmap(self.embed)(sequence_tokens)
        x, _, hiddens = self.transformer(x, sequence_id=sequence_id)
        hiddens = jnp.stack(hiddens, axis=0)
        sequence_logits = x
        for l in self.sequence_head:
            sequence_logits = eqx.filter_vmap(l)(sequence_logits)

        return sequence_logits, x, hiddens

    @staticmethod
    def from_pretrained(
        model: Literal["esmc_300m", "esmc_600m"],
        *,
        key: PRNGKeyArray | None = None,
        dtype: Any | None = None,
        inference: bool = False,
    ) -> "ESMC":
        import torch

        if model == "esmc_300m":
            model_path = data_root("esmc-300") / "data/weights/esmc_300m_2024_12_v0.pth"
            d_model = 960
            n_heads = 15
            n_layers = 30

        elif model == "esmc_600m":
            d_model = 1152
            n_heads = 18
            n_layers = 36

            model_path = data_root("esmc-600") / "data/weights/esmc_600m_2024_12_v0.pth"
        else:
            raise ValueError(
                f"Received invalid model: {model} It must be esmc_300m or esmc_600m"
            )
        key = jax.random.key(42) if key is None else key
        model_path = cast(Path, model_path)
        state_dict = torch.load(model_path, map_location="cpu")

        esmc = ESMC(
            d_model=d_model,
            n_heads=n_heads,
            n_layers=n_layers,
            key=key,
            dtype=dtype,
            inference=inference,
        )

        esmc = autoconvert(esmc, state_dict)

        return esmc


def rbf(values, v_min, v_max, n_bins=16):
    rbf_centers = jnp.linspace(v_min, v_max, n_bins, dtype=values.dtype)
    rbf_centers = rbf_centers.reshape([1] * len(values.shape) + [-1])
    rbf_std = (v_max - v_min) / n_bins
    z = (jnp.expand_dims(values, axis=-1) - rbf_centers) / rbf_std
    return jnp.exp(-(z**2))


def build_affine3d_from_coordinates(
    coords: Array,
) -> tuple[Affine3D, Array]:
    _MAX_SUPPORTED_DISTANCE = 1e6
    coord_mask = jnp.all(
        jnp.all(jnp.isfinite(coords) & (coords < _MAX_SUPPORTED_DISTANCE), axis=-1),
        axis=-1,
    )

    def atom3_to_backbone_affine(bb_positions: Array) -> Affine3D:
        N = bb_positions[..., 0, :]
        CA = bb_positions[..., 1, :]
        C = bb_positions[..., 2, :]
        return Affine3D.from_graham_schmidt(C, CA, N)

    coords = coords.astype(jnp.float32)
    coords = jnp.where(coord_mask[:, None, None], coords, 0.0)

    masked_coords = jnp.where(coord_mask[:, None, None], coords, 0.0)
    average_per_n_ca_c = masked_coords.sum(axis=0) / (coord_mask.sum() + 1e-8)

    affine_from_average = atom3_to_backbone_affine(average_per_n_ca_c).as_matrix()

    S = coords.shape[0]

    affine_rot_mats = jnp.broadcast_to(affine_from_average.rot.tensor[None, :], (S, 9))
    affine_trans = jnp.broadcast_to(affine_from_average.trans[None, :], (S, 3))

    identity_rot = jax.lax.stop_gradient(
        RotationMatrix.identity((S,), dtype=jnp.float32)
    )

    has_any_coords = coord_mask.any()
    affine_rot_mats = jnp.where(has_any_coords, affine_rot_mats, identity_rot.tensor)

    black_hole_affine = Affine3D(affine_trans, RotationMatrix(affine_rot_mats))

    affine = atom3_to_backbone_affine(coords)
    affine = Affine3D.from_tensor(
        jnp.where(coord_mask[:, None], affine.tensor, black_hole_affine.tensor)
    )

    return affine, coord_mask


class OutputHeads(eqx.Module):
    sequence_head: list
    structure_head: list
    ss8_head: list
    sasa_head: list
    function_head: list
    residue_head: list

    def __init__(
        self,
        d_model: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if not dtype:
            dtype = default_floating_dtype()
        assert dtype is not None
        keys = jax.random.split(key, 6)
        self.sequence_head = RegressionHead(d_model, 64, key=keys[0], dtype=dtype)
        self.structure_head = RegressionHead(d_model, 4096, key=keys[1], dtype=dtype)
        self.ss8_head = RegressionHead(d_model, 8 + 3, key=keys[2], dtype=dtype)
        self.sasa_head = RegressionHead(d_model, 16 + 3, key=keys[3], dtype=dtype)
        self.function_head = RegressionHead(d_model, 260 * 8, key=keys[4], dtype=dtype)
        self.residue_head = RegressionHead(d_model, 1478, key=keys[5], dtype=dtype)

    def _apply_head(self, head: list, x: Array) -> Array:
        for layer in head:
            x = eqx.filter_vmap(layer)(x)
        return x

    def __call__(
        self, x: Array, embed: Array
    ) -> tuple[Array, Array, Array, Array, Array, Array, Array]:
        sequence_logits = self._apply_head(self.sequence_head, x)
        structure_logits = self._apply_head(self.structure_head, x)
        secondary_structure_logits = self._apply_head(self.ss8_head, x)
        sasa_logits = self._apply_head(self.sasa_head, x)
        function_logits = self._apply_head(self.function_head, x)
        function_logits = function_logits.reshape(*function_logits.shape[:-1], 8, -1)
        residue_logits = self._apply_head(self.residue_head, x)
        return (
            sequence_logits,
            structure_logits,
            secondary_structure_logits,
            sasa_logits,
            function_logits,
            residue_logits,
            embed,
        )


class EncodeInputs(eqx.Module):
    """
    Module for encoding input features in the ESM-3 model.

    Args:
        d_model (int): The dimensionality of the model's hidden states.
    """

    sequence_embed: eqx.nn.Embedding
    plddt_projection: eqx.nn.Linear
    structure_per_res_plddt_projection: eqx.nn.Linear
    structure_tokens_embed: eqx.nn.Embedding
    ss8_embed: eqx.nn.Embedding
    sasa_embed: eqx.nn.Embedding
    function_embed: list[EmbeddingWithPadding]
    residue_embed: EmbeddingBag

    def __init__(
        self,
        d_model: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if not dtype:
            dtype = default_floating_dtype()
        assert dtype is not None

        key, seq_embed_key = jax.random.split(key)
        self.sequence_embed = eqx.nn.Embedding(
            64, d_model, key=seq_embed_key, dtype=dtype
        )
        key, plddt_projection_key, struc_key = jax.random.split(key, 3)
        self.plddt_projection = eqx.nn.Linear(
            16, d_model, key=plddt_projection_key, dtype=dtype
        )
        self.structure_per_res_plddt_projection = eqx.nn.Linear(
            16, d_model, key=struc_key, dtype=dtype
        )

        key, struc_embed_key = jax.random.split(key)
        self.structure_tokens_embed = eqx.nn.Embedding(
            4096 + 5, d_model, key=struc_embed_key, dtype=dtype
        )

        key, ss8_key, sasa_key = jax.random.split(key, 3)
        # "Structural" features
        self.ss8_embed = eqx.nn.Embedding(8 + 3, d_model, key=ss8_key, dtype=dtype)
        self.sasa_embed = eqx.nn.Embedding(16 + 3, d_model, key=sasa_key, dtype=dtype)

        # "Functional" features
        key, *func_keys = jax.random.split(key, 9)
        self.function_embed = [
            EmbeddingWithPadding(
                260, d_model // 8, padding_idx=0, key=func_keys[i], dtype=dtype
            )
            for i in range(8)
        ]

        key, residue_key = jax.random.split(key)
        self.residue_embed = EmbeddingBag(
            1478, d_model, padding_idx=0, key=residue_key, dtype=dtype
        )

    def __call__(
        self,
        sequence_tokens: Array,
        structure_tokens: Array,
        average_plddt: Array,
        per_res_plddt: Array,
        ss8_tokens: Array,
        sasa_tokens: Array,
        function_tokens: Array,
        residue_annotation_tokens: Array,
    ) -> Array:
        sequence_embed = eqx.filter_vmap(self.sequence_embed)(sequence_tokens)

        rbf_16_fn = functools.partial(rbf, v_min=0.0, v_max=1.0, n_bins=16)
        plddt_embed = self.plddt_projection(rbf_16_fn(average_plddt))
        structure_per_res_plddt = eqx.filter_vmap(
            self.structure_per_res_plddt_projection
        )(rbf_16_fn(per_res_plddt))

        structure_embed = eqx.filter_vmap(self.structure_tokens_embed)(structure_tokens)
        ss8_embed = eqx.filter_vmap(self.ss8_embed)(ss8_tokens)
        sasa_embed = eqx.filter_vmap(self.sasa_embed)(sasa_tokens)

        function_embed = jnp.concatenate(
            [
                eqx.filter_vmap(embed_fn)(function_tokens[..., i])
                for i, embed_fn in enumerate(self.function_embed)
            ],
            axis=-1,
        )

        residue_embed = eqx.filter_vmap(self.residue_embed)(residue_annotation_tokens)

        return (
            sequence_embed
            + plddt_embed
            + structure_per_res_plddt
            + structure_embed
            + ss8_embed
            + sasa_embed
            + function_embed
            + residue_embed
        )


class ESM3(eqx.Module):
    encoder: EncodeInputs
    transformer: TransformerStack
    output_heads: OutputHeads

    inference: bool

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        v_heads: int,
        n_layers: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        inference: bool = False,
    ):
        if not dtype:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.inference = inference

        key, enc_key, trans_key, head_key = jax.random.split(key, 4)
        self.encoder = EncodeInputs(d_model, key=enc_key, dtype=dtype)
        self.transformer = TransformerStack(
            d_model,
            n_heads,
            v_heads,
            n_layers,
            mask_and_zero_frameless=True,
            key=trans_key,
            dtype=dtype,
            inference=inference,
        )
        self.output_heads = OutputHeads(d_model, key=head_key, dtype=dtype)

    def __call__(
        self,
        sequence_tokens: Array | None = None,
        structure_tokens: Array | None = None,
        ss8_tokens: Array | None = None,
        sasa_tokens: Array | None = None,
        function_tokens: Array | None = None,
        residue_annotation_tokens: Array | None = None,
        average_plddt: Array | None = None,
        per_res_plddt: Array | None = None,
        structure_coords: Array | None = None,
        chain_id: Array | None = None,
        sequence_id: Array | None = None,
    ) -> tuple[Array, Array, Array, Array, Array, Array, Array]:

        L = next(
            x.shape[0]
            for x in [
                sequence_tokens,
                structure_tokens,
                ss8_tokens,
                sasa_tokens,
                structure_coords,
                function_tokens,
                residue_annotation_tokens,
            ]
            if x is not None
        )

        if sequence_tokens is None:
            sequence_tokens = jnp.full((L,), SEQUENCE_MASK_TOKEN, dtype=jnp.int32)
        if ss8_tokens is None:
            ss8_tokens = jnp.full((L,), SS8_PAD_TOKEN, dtype=jnp.int32)
        if sasa_tokens is None:
            sasa_tokens = jnp.full((L,), SASA_PAD_TOKEN, dtype=jnp.int32)
        if average_plddt is None:
            average_plddt = jnp.float32(1.0)
        if per_res_plddt is None:
            per_res_plddt = jnp.zeros((L,), dtype=jnp.float32)
        if chain_id is None:
            chain_id = jnp.zeros((L,), dtype=jnp.int32)
        if residue_annotation_tokens is None:
            residue_annotation_tokens = jnp.full(
                (L, 16), RESIDUE_PAD_TOKEN, dtype=jnp.int32
            )
        if function_tokens is None:
            function_tokens = jnp.full((L, 8), INTERPRO_PAD_TOKEN, dtype=jnp.int32)
        if structure_coords is None:
            structure_coords = jnp.full((L, 3, 3), jnp.nan, dtype=jnp.float32)

        structure_coords = structure_coords[..., :3, :]
        affine, affine_mask = build_affine3d_from_coordinates(structure_coords)

        if structure_tokens is None:
            structure_tokens = jnp.full((L,), STRUCTURE_MASK_TOKEN, dtype=jnp.int32)

        structure_tokens = jnp.where(
            structure_tokens == -1, STRUCTURE_MASK_TOKEN, structure_tokens
        )
        structure_tokens = jnp.where(
            sequence_tokens == SEQUENCE_BOS_TOKEN, STRUCTURE_BOS_TOKEN, structure_tokens
        )
        structure_tokens = jnp.where(
            sequence_tokens == SEQUENCE_PAD_TOKEN, STRUCTURE_PAD_TOKEN, structure_tokens
        )
        structure_tokens = jnp.where(
            sequence_tokens == SEQUENCE_EOS_TOKEN, STRUCTURE_EOS_TOKEN, structure_tokens
        )
        structure_tokens = jnp.where(
            sequence_tokens == SEQUENCE_CHAINBREAK_TOKEN,
            STRUCTURE_CHAINBREAK_TOKEN,
            structure_tokens,
        )

        x = self.encoder(
            sequence_tokens,
            structure_tokens,
            average_plddt,
            per_res_plddt,
            ss8_tokens,
            sasa_tokens,
            function_tokens,
            residue_annotation_tokens,
        )

        x, embedding, _ = self.transformer(
            x, sequence_id, affine, affine_mask, chain_id
        )

        return self.output_heads(x, embedding)

    @staticmethod
    def from_pretrained(
        model: Literal["esm3_open"] = "esm3_open",
        *,
        key: PRNGKeyArray | None = None,
        dtype: Any | None = None,
        inference: bool = False,
    ) -> "ESM3":
        import torch

        if model == "esm3_open":
            d_model = 1536
            n_heads = 24
            v_heads = 256
            n_layers = 48
            model_path = data_root("esm3") / "data/weights/esm3_sm_open_v1.pth"
        else:
            raise ValueError(f"Unknown model: {model}. Must be 'esm3_open'.")

        key = jax.random.key(42) if key is None else key
        state_dict = torch.load(model_path, map_location="cpu")
        state_dict = {
            k: v.float() if v.dtype == torch.bfloat16 else v
            for k, v in state_dict.items()
        }
        esm3 = ESM3(
            d_model=d_model,
            n_heads=n_heads,
            v_heads=v_heads,
            n_layers=n_layers,
            key=key,
            dtype=dtype,
            inference=inference,
        )
        esm3 = autoconvert(esm3, state_dict)
        return esm3


BB_COORDINATES = np.array([[0.5256, 1.3612, 0.0], [0.0, 0.0, 0.0], [-1.5251, 0.0, 0.0]])


class Dim6RotStructureHead(eqx.Module):
    ffn1: eqx.nn.Linear
    activation_fn: eqx.nn.Lambda
    norm: eqx.nn.LayerNorm
    proj: eqx.nn.Linear
    trans_scale_factor: float = eqx.field(static=True)

    def __init__(
        self,
        input_dim: int,
        trans_scale_factor: float = 10,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        key1, key2 = jax.random.split(key)
        self.ffn1 = eqx.nn.Linear(input_dim, input_dim, key=key1, dtype=dtype)
        self.activation_fn = eqx.nn.Lambda(fn=jax.nn.gelu)
        self.norm = eqx.nn.LayerNorm(input_dim)
        self.proj = eqx.nn.Linear(input_dim, 9 + 7 * 2, key=key2, dtype=dtype)
        self.trans_scale_factor = trans_scale_factor

    def __call__(
        self, x: Array, affine: Affine3D | None, affine_mask: Array
    ) -> tuple[Array, Array]:
        seq_len = x.shape[0]

        if affine is None:
            rigids = Affine3D.identity(
                (seq_len,),
                rotation_type=RotationMatrix,
            )
        else:
            rigids = affine

        x = eqx.filter_vmap(self.ffn1)(x)
        x = jax.nn.gelu(x)
        x = eqx.filter_vmap(self.norm)(x)
        x = eqx.filter_vmap(self.proj)(x)

        trans, x_axis, y_axis, angles = jnp.split(x, [3, 6, 9], axis=-1)
        trans = trans * self.trans_scale_factor
        x_axis = x_axis / (jnp.linalg.norm(x_axis, axis=-1, keepdims=True) + 1e-5)
        y_axis = y_axis / (jnp.linalg.norm(y_axis, axis=-1, keepdims=True) + 1e-5)

        update = Affine3D.from_graham_schmidt(x_axis + trans, trans, y_axis + trans)
        update = update.mask(affine_mask)
        rigids = rigids.compose(update)
        affine_tensor = rigids.tensor

        bb_local = jnp.broadcast_to(
            jnp.array(BB_COORDINATES)[None, :, :], (seq_len, 3, 3)
        )
        pred_xyz = rigids[..., None].apply(bb_local)

        return affine_tensor, pred_xyz


class PairwisePredictionHead(eqx.Module):
    downproject: eqx.nn.Linear
    linear1: eqx.nn.Linear
    norm: eqx.nn.LayerNorm
    linear2: eqx.nn.Linear

    def __init__(
        self,
        input_dim: int,
        downproject_dim: int,
        hidden_dim: int,
        n_bins: int,
        bias: bool = True,
        pairwise_state_dim: int = 0,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        k1, k2, k3 = jax.random.split(key, 3)
        self.downproject = eqx.nn.Linear(
            input_dim, downproject_dim, use_bias=bias, key=k1, dtype=dtype
        )
        self.linear1 = eqx.nn.Linear(
            downproject_dim + pairwise_state_dim,
            hidden_dim,
            use_bias=bias,
            key=k2,
            dtype=dtype,
        )
        self.norm = eqx.nn.LayerNorm(hidden_dim)
        self.linear2 = eqx.nn.Linear(
            hidden_dim, n_bins, use_bias=bias, key=k3, dtype=dtype
        )

    def __call__(self, x: Array, pairwise: Array | None = None) -> Array:
        x = eqx.filter_vmap(self.downproject)(x)
        half = x.shape[-1] // 2
        q, k = x[..., :half], x[..., half:]

        prod = q[None, :, :] * k[:, None, :]
        diff = q[None, :, :] - k[:, None, :]

        if pairwise is not None:
            x = jnp.concatenate([prod, diff, pairwise], axis=-1)
        else:
            x = jnp.concatenate([prod, diff], axis=-1)

        apply_linear1 = eqx.filter_vmap(eqx.filter_vmap(self.linear1))
        x = apply_linear1(x)
        x = jax.nn.gelu(x)
        apply_norm = eqx.filter_vmap(eqx.filter_vmap(self.norm))
        x = apply_norm(x)
        apply_linear2 = eqx.filter_vmap(eqx.filter_vmap(self.linear2))
        x = apply_linear2(x)

        return x


class VQVAERegressionHead(eqx.Module):
    dense: eqx.nn.Linear
    norm: eqx.nn.LayerNorm
    output: eqx.nn.Linear

    def __init__(
        self,
        embed_dim: int,
        output_dim: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        k1, k2 = jax.random.split(key)
        self.dense = eqx.nn.Linear(embed_dim, embed_dim, key=k1, dtype=dtype)
        self.norm = eqx.nn.LayerNorm(embed_dim)
        self.output = eqx.nn.Linear(embed_dim, output_dim, key=k2, dtype=dtype)

    def __call__(self, x: Array) -> Array:
        x = eqx.filter_vmap(self.dense)(x)
        x = jax.nn.gelu(x)
        x = eqx.filter_vmap(self.norm)(x)
        x = eqx.filter_vmap(self.output)(x)
        return x


VQVAE_CODEBOOK_SIZE = 4096
VQVAE_SPECIAL_TOKENS = {
    "MASK": 4096,
    "EOS": 4097,
    "BOS": 4098,
    "PAD": 4099,
    "CHAINBREAK": 4100,
}
VQVAE_DIRECTION_LOSS_BINS = 16
VQVAE_PAE_BINS = 64
VQVAE_MAX_PAE_BIN = 31.0
VQVAE_PLDDT_BINS = 50


def compute_predicted_aligned_error(
    logits: Array,
    aa_mask: Array,
    sequence_id: Array | None = None,
    max_bin: float = 31.0,
) -> Array:
    n_bins = logits.shape[-1]
    bins = jnp.linspace(0, max_bin, n_bins - 1)
    step = max_bin / (n_bins - 2)
    bin_centers = bins + step / 2
    bin_centers = jnp.concatenate([bin_centers, jnp.array([bin_centers[-1] + step])])
    square_mask = aa_mask[:, None] * aa_mask[None, :]
    min_v = jnp.finfo(logits.dtype).min
    probs = jax.nn.softmax(jnp.where(square_mask[..., None], logits, min_v), axis=-1)
    return (probs * bin_centers).sum(axis=-1)


def compute_tm(
    logits: Array,
    aa_mask: Array,
    max_bin: float = 31.0,
) -> Array:
    square_mask = (aa_mask[:, None] * aa_mask[None, :]).astype(jnp.bool_)
    seqlen = aa_mask.sum()
    n_bins = logits.shape[-1]
    bins = jnp.linspace(0, max_bin, n_bins - 1)
    step = max_bin / (n_bins - 2)
    bin_centers = bins + step / 2
    bin_centers = jnp.concatenate([bin_centers, jnp.array([bin_centers[-1] + step])])
    d0 = 1.24 * (jnp.maximum(seqlen, 19) - 15) ** (1.0 / 3.0) - 1.8
    f_d = 1.0 / (1 + (bin_centers / d0) ** 2)
    min_v = jnp.finfo(logits.dtype).min
    probs = jax.nn.softmax(jnp.where(square_mask[..., None], logits, min_v), axis=-1)
    ptm = (probs * f_d[None, :]).sum(axis=-1)
    eps = 1e-10
    ptm = jnp.where(square_mask, ptm, 0.0).sum(axis=-1) / (
        square_mask.sum(axis=-1) + eps
    )
    return ptm.max()


def categorical_mixture_mean(
    logits: Array, n_bins: int, start: float = 0.0, end: float = 1.0
) -> Array:
    bins = jnp.linspace(start, end, n_bins + 1)
    v_bins = (bins[:-1] + bins[1:]) / 2
    probs = jax.nn.softmax(logits, axis=-1)
    return (probs * v_bins).sum(axis=-1)


class StructureTokenDecoder(eqx.Module):
    embed: eqx.nn.Embedding
    decoder_stack: TransformerStack
    affine_output_projection: Dim6RotStructureHead
    pairwise_classification_head: PairwisePredictionHead
    plddt_head: VQVAERegressionHead

    pairwise_bins: list[int] = eqx.field(static=True)

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        n_layers: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        k1, k2, k3, k4, k5 = jax.random.split(key, 5)

        self.embed = eqx.nn.Embedding(
            VQVAE_CODEBOOK_SIZE + len(VQVAE_SPECIAL_TOKENS),
            d_model,
            key=k1,
            dtype=dtype,
        )
        self.decoder_stack = TransformerStack(
            d_model,
            n_heads,
            1,
            n_layers,
            scale_residue=False,
            n_layers_geom=0,
            key=k2,
            dtype=dtype,
        )
        self.affine_output_projection = Dim6RotStructureHead(
            d_model, trans_scale_factor=10, key=k3, dtype=dtype
        )

        self.pairwise_bins = [64, VQVAE_DIRECTION_LOSS_BINS * 6, VQVAE_PAE_BINS]
        self.pairwise_classification_head = PairwisePredictionHead(
            input_dim=d_model,
            downproject_dim=128,
            hidden_dim=128,
            n_bins=sum(self.pairwise_bins),
            bias=False,
            key=k4,
            dtype=dtype,
        )
        self.plddt_head = VQVAERegressionHead(
            embed_dim=d_model, output_dim=VQVAE_PLDDT_BINS, key=k5, dtype=dtype
        )

    def decode(
        self,
        structure_tokens: Array,
        sequence_id: Array | None = None,
    ) -> dict:
        seq_len = structure_tokens.shape[0]

        if sequence_id is None:
            sequence_id = jnp.zeros(seq_len, dtype=jnp.int32)
        chain_id = jnp.zeros(seq_len, dtype=jnp.int32)

        x = eqx.filter_vmap(self.embed)(structure_tokens)
        x, _, _ = self.decoder_stack(
            x, sequence_id=sequence_id, frames=None, frames_mask=None, chain_id=chain_id
        )

        affine_mask = jnp.zeros(seq_len, dtype=jnp.bool_)
        tensor7_affine, bb_pred = self.affine_output_projection(
            x, affine=None, affine_mask=affine_mask
        )

        pairwise_logits = self.pairwise_classification_head(x)
        _, _, pae_logits = jnp.split(
            pairwise_logits,
            [self.pairwise_bins[0], self.pairwise_bins[0] + self.pairwise_bins[1]],
            axis=-1,
        )

        special_tokens_mask = structure_tokens >= min(VQVAE_SPECIAL_TOKENS.values())
        aa_mask = ~special_tokens_mask

        pae = compute_predicted_aligned_error(
            pae_logits, aa_mask, sequence_id=sequence_id, max_bin=VQVAE_MAX_PAE_BIN
        )
        ptm = compute_tm(pae_logits, aa_mask, max_bin=VQVAE_MAX_PAE_BIN)

        plddt_logits = self.plddt_head(x)
        plddt_value = categorical_mixture_mean(
            plddt_logits, n_bins=plddt_logits.shape[-1]
        )

        return {
            "tensor7_affine": tensor7_affine,
            "bb_pred": bb_pred,
            "plddt": plddt_value,
            "ptm": ptm,
            "predicted_aligned_error": pae,
        }

    @staticmethod
    def from_pretrained(
        *,
        key: PRNGKeyArray | None = None,
        dtype: Any | None = None,
    ) -> "StructureTokenDecoder":
        import torch

        d_model = 1280
        n_heads = 20
        n_layers = 30
        model_path = data_root("esm3") / "data/weights/esm3_structure_decoder_v0.pth"

        key = jax.random.key(42) if key is None else key
        state_dict = torch.load(model_path, map_location="cpu")
        state_dict = {
            k: v.float() if v.dtype == torch.bfloat16 else v
            for k, v in state_dict.items()
        }

        decoder = StructureTokenDecoder(
            d_model=d_model,
            n_heads=n_heads,
            n_layers=n_layers,
            key=key,
            dtype=dtype,
        )
        decoder = autoconvert(decoder, state_dict)
        return decoder


def cosine_schedule(t):
    return jnp.cos(t * math.pi * 0.5)


def sample_structure_tokens(
    logits: Array,
    temperature: float = 1.0,
    top_p: float = 1.0,
    *,
    key: PRNGKeyArray,
) -> Array:
    if temperature <= 0:
        return jnp.argmax(logits, axis=-1)

    logits = logits / temperature

    if top_p < 1.0:
        sorted_indices = jnp.argsort(logits, axis=-1)[::-1]
        sorted_logits = jnp.take_along_axis(logits, sorted_indices, axis=-1)
        sorted_probs = jax.nn.softmax(sorted_logits, axis=-1)
        cumulative_probs = jnp.cumsum(sorted_probs, axis=-1)
        mask = cumulative_probs - sorted_probs > top_p
        sorted_logits = jnp.where(mask, -jnp.inf, sorted_logits)
        logits = jnp.zeros_like(logits).at[sorted_indices].set(sorted_logits)

    return jax.random.categorical(key, logits, axis=-1)


def get_annealed_temperature(step: int, num_steps: int, initial_temperature: float):
    step_ratio = step / max(1, num_steps - 1)
    return max(initial_temperature - step_ratio, 0.001) ** 2


def _sample_step(
    structure_logits: Array,
    structure_tokens: Array,
    can_sample: Array,
    total_to_sample: int,
    step: int,
    num_steps: int,
    temperature: float,
    top_p: float,
    strategy: str,
    key: PRNGKeyArray,
) -> Array:
    sample_key, perm_key = jax.random.split(key)
    seq_len = structure_tokens.shape[0]

    sampled = sample_structure_tokens(
        structure_logits[:, :4096],
        temperature=temperature,
        top_p=top_p,
        key=sample_key,
    )

    still_masked_count = can_sample.sum()
    perc_masked_after = cosine_schedule(jnp.array((step + 1) / num_steps))
    num_masked_after = jnp.int32(perc_masked_after * total_to_sample + 0.1)
    num_to_sample = jnp.maximum(still_masked_count - num_masked_after, 0)

    if strategy == "entropy":
        logprobs = jax.nn.log_softmax(structure_logits[:, :4096], axis=-1)
        probs = jnp.exp(logprobs)
        entropy = -(probs * logprobs).sum(axis=-1)
        scores = jnp.where(can_sample, entropy, jnp.inf)
    else:
        scores = jnp.where(
            can_sample, jax.random.uniform(perm_key, (seq_len,)), jnp.inf
        )

    ranks = jnp.argsort(jnp.argsort(scores))
    where_to_sample = (ranks < num_to_sample) & can_sample

    return jnp.where(where_to_sample, sampled, structure_tokens)


def iterative_sample_structure(
    model,
    sequence_tokens: Array,
    num_steps: int = 8,
    temperature: float = 1.0,
    temperature_annealing: bool = True,
    top_p: float = 1.0,
    strategy: str = "random",
    *,
    key: PRNGKeyArray,
) -> Array:
    seq_len = sequence_tokens.shape[0]

    structure_tokens = jnp.full(seq_len, STRUCTURE_MASK_TOKEN, dtype=jnp.int32)
    structure_tokens = structure_tokens.at[0].set(STRUCTURE_BOS_TOKEN)
    structure_tokens = structure_tokens.at[-1].set(STRUCTURE_EOS_TOKEN)

    fixed_mask = jnp.ones(seq_len, dtype=jnp.bool_)
    fixed_mask = fixed_mask.at[0].set(False)
    fixed_mask = fixed_mask.at[-1].set(False)

    total_to_sample = int(fixed_mask.sum())

    for t in range(num_steps):
        key, step_key = jax.random.split(key)

        output = model(
            sequence_tokens=sequence_tokens, structure_tokens=structure_tokens
        )
        structure_logits = output[1]

        if temperature_annealing:
            temp = get_annealed_temperature(t, num_steps, temperature)
        else:
            temp = temperature

        can_sample = fixed_mask & (structure_tokens == STRUCTURE_MASK_TOKEN)

        structure_tokens = _sample_step(
            structure_logits,
            structure_tokens,
            can_sample,
            total_to_sample,
            t,
            num_steps,
            temp,
            top_p,
            strategy,
            step_key,
        )

    return structure_tokens


def tokenize_sequence(sequence: str) -> list[int]:
    tokens = [SEQUENCE_BOS_TOKEN]  # BOS = <cls>
    for ch in sequence:
        tokens.append(SEQUENCE_VOCAB.index(ch))
    tokens.append(SEQUENCE_EOS_TOKEN)  # EOS = <eos>
    return tokens


def fold_protein(
    esm3_model,
    decoder: StructureTokenDecoder,
    sequence: str,
    num_steps: int = 8,
    temperature: float = 1.0,
    temperature_annealing: bool = True,
    top_p: float = 1.0,
    strategy: str = "random",
    *,
    key: PRNGKeyArray,
) -> dict:
    tokens = tokenize_sequence(sequence)
    sequence_tokens = jnp.array(tokens, dtype=jnp.int32)

    structure_tokens = iterative_sample_structure(
        esm3_model,
        sequence_tokens,
        num_steps=num_steps,
        temperature=temperature,
        temperature_annealing=temperature_annealing,
        top_p=top_p,
        strategy=strategy,
        key=key,
    )

    result = decoder.decode(structure_tokens)

    return {
        "sequence": sequence,
        "structure_tokens": structure_tokens,
        "bb_coords": result["bb_pred"],
        "plddt": result["plddt"],
        "ptm": result["ptm"],
        "pae": result["predicted_aligned_error"],
    }



import functools

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from jaxonlayers.layers import LayerNorm
from statedict2pytree import (
    convert,
    pytree_to_fields,
    state_dict_to_fields,
)
from torchvision.models.swin_transformer import PatchMerging as TorchPatchMerging
from torchvision.models.swin_transformer import PatchMergingV2 as TorchPatchMergingV2
from torchvision.models.swin_transformer import (
    ShiftedWindowAttention as TorchShiftedWindowAttentionV1,
)
from torchvision.models.swin_transformer import (
    ShiftedWindowAttentionV2 as TorchShiftedWindowAttentionV2,
)
from torchvision.models.swin_transformer import SwinTransformer as TorchSwinTransformer
from torchvision.models.swin_transformer import (
    SwinTransformerBlock as TorchSwinTransformerBlock,
)
from torchvision.models.swin_transformer import (
    SwinTransformerBlockV2 as TorchSwinTransformerBlockV2,
)

from jaxonmodels.functions import default_floating_dtype
from jaxonmodels.models.swin_transformer import (
    PatchMerging,
    PatchMergingV2,
    ShiftedWindowAttention,
    ShiftedWindowAttentionV2,
    SwinTransformer,
    SwinTransformerBlock,
    SwinTransformerBlockV2,
)
from jaxonmodels.statedict2pytree import model_orders


@pytest.mark.parametrize(
    "C, H, W",
    [
        (32, 55, 55),
        (32, 224, 224),
        (16, 56, 56),
        (32, 112, 56),
    ],
)
def test_patch_merging_layers(C, H, W):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_patch = TorchPatchMerging(dim=C)
    jax_patch, state = eqx.nn.make_with_state(PatchMerging)(
        dim=C,
        norm_layer=functools.partial(LayerNorm, dtype=dtype),
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_patch, state = eqx.nn.inference_mode((jax_patch, state))

    weights_dict = torch_patch.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    jaxfields, state_indices = pytree_to_fields((jax_patch, state))

    jax_patch, state = convert(
        weights_dict,
        (jax_patch, state),
        jaxfields,
        state_indices,
        torchfields,
    )
    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_patch(torch.from_numpy(x))
    j_out, state = eqx.filter_vmap(jax_patch, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-6)


@pytest.mark.parametrize(
    "C, window_size, shift_size, num_heads, qkv_bias, proj_bias, attention_dropout, dropout, H, W",  # noqa
    [
        (96, [7, 7], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [7, 7], [3, 3], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [7, 7], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [7, 7], [3, 3], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [7, 7], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [7, 7], [3, 3], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [7, 7], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [7, 7], [3, 3], 24, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [4, 4], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [4, 4], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [4, 4], 24, True, True, 0.0, 0.0, 56, 56),
    ],
)
def test_swin_attention_v1(
    C,
    window_size,
    shift_size,
    num_heads,
    qkv_bias,
    proj_bias,
    attention_dropout,
    dropout,
    H,
    W,
):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_swin = TorchShiftedWindowAttentionV1(C, window_size, shift_size, num_heads)

    jax_swin, state = eqx.nn.make_with_state(ShiftedWindowAttention)(
        dim=C,
        shift_size=shift_size,
        num_heads=num_heads,
        qkv_bias=qkv_bias,
        window_size=window_size,
        proj_bias=proj_bias,
        attention_dropout=attention_dropout,
        dropout=dropout,
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    # torchfields = move_running_fields_to_the_end(
    #     torchfields, identifier="relative_position_index"
    # )
    jaxfields, state_indices = pytree_to_fields((jax_swin, state))

    # for t, j in zip(torchfields, jaxfields):
    #     print(t.path, t.shape, jax.tree_util.keystr(j.path), j.shape)

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_swin(torch.from_numpy(x))

    j_out, state = eqx.filter_vmap(jax_swin, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-5)


@pytest.mark.parametrize(
    "C, window_size, shift_size, num_heads, qkv_bias, proj_bias, attention_dropout, dropout, H, W",  # noqa
    [
        (96, [8, 8], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [4, 4], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [4, 4], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [4, 4], 24, True, True, 0.0, 0.0, 56, 56),
    ],
)
def test_swin_attention_v2(
    C,
    window_size,
    shift_size,
    num_heads,
    qkv_bias,
    proj_bias,
    attention_dropout,
    dropout,
    H,
    W,
):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_swin = TorchShiftedWindowAttentionV2(C, window_size, shift_size, num_heads)

    jax_swin, state = eqx.nn.make_with_state(ShiftedWindowAttentionV2)(
        dim=C,
        shift_size=shift_size,
        num_heads=num_heads,
        qkv_bias=qkv_bias,
        window_size=window_size,
        proj_bias=proj_bias,
        attention_dropout=attention_dropout,
        dropout=dropout,
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    # torchfields = move_running_fields_to_the_end(
    #     torchfields, identifier="relative_position_index"
    # )
    # torchfields = move_running_fields_to_the_end(
    #     torchfields, identifier="relative_coords_table"
    # )

    jaxfields, state_indices = pytree_to_fields((jax_swin, state))

    for t, j in zip(torchfields, jaxfields):
        print(t.path, t.shape, jax.tree_util.keystr(j.path), j.shape)

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_swin(torch.from_numpy(x))

    j_out, state = eqx.filter_vmap(jax_swin, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-5)


@pytest.mark.parametrize(
    "dim, num_heads, window_size, shift_size, mlp_ratio, qkv_bias, proj_bias, dropout, attention_dropout, stochastic_depth_prob, H, W",  # noqa
    [
        (96, 3, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (96, 3, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (192, 6, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 28, 28),
        (192, 6, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 28, 28),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (768, 24, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 7, 7),
        (768, 24, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0, 7, 7),
    ],
)
def test_swin_transformer_block(
    dim,
    num_heads,
    window_size,
    shift_size,
    mlp_ratio,
    qkv_bias,
    proj_bias,
    dropout,
    attention_dropout,
    stochastic_depth_prob,
    H,
    W,
):
    batch_size = 2
    dtype = default_floating_dtype()
    key = jax.random.key(43)
    torch.manual_seed(42)
    np.random.seed(42)

    # --- PyTorch Model ---

    torch_block = TorchSwinTransformerBlock(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=torch.nn.LayerNorm,
        attn_layer=TorchShiftedWindowAttentionV1,
    )
    torch_block.eval()

    # --- JAX Model ---
    #  Always use the V1 attention module.
    jax_block, state = eqx.nn.make_with_state(SwinTransformerBlock)(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=functools.partial(eqx.nn.LayerNorm, eps=1e-05),  # pyright: ignore
        attn_layer=ShiftedWindowAttention,
        inference=False,
        dtype=dtype,
        key=key,
    )
    jax_block, state = eqx.nn.inference_mode((jax_block, state))

    # --- Weight Conversion ---
    weights_dict = torch_block.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    # torchfields = move_running_fields_to_the_end(
    #     torchfields,
    #     identifier="relative_position_index",
    # )

    jaxfields, state_indices = pytree_to_fields(
        (
            jax_block,
            state,
        )
    )

    jax_block, state = convert(
        weights_dict,
        (jax_block, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    # --- Inference and Comparison ---
    x_np = np.array(np.random.uniform(size=(batch_size, H, W, dim)), dtype=np.float32)
    x_torch = torch.from_numpy(x_np)
    x_jax = jnp.array(x_np)

    t_out = torch_block(x_torch)
    jax_block_pt = functools.partial(jax_block, key=jax.random.key(2100))
    j_out, state = eqx.filter_vmap(jax_block_pt, in_axes=(0, None), out_axes=(0, None))(
        x_jax, state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-3)


@pytest.mark.parametrize(
    "dim, num_heads, window_size, shift_size, mlp_ratio, qkv_bias, proj_bias, dropout, attention_dropout, stochastic_depth_prob, H, W",  # noqa
    [
        (96, 3, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (96, 3, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (192, 6, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 28, 28),
        (192, 6, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.0, 28, 28),
        (384, 12, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (384, 12, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.0, 14, 14),
        (768, 24, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 7, 7),
        (768, 24, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.0, 7, 7),
    ],
)
def test_swin_transformer_blockv2(
    dim,
    num_heads,
    window_size,
    shift_size,
    mlp_ratio,
    qkv_bias,
    proj_bias,
    dropout,
    attention_dropout,
    stochastic_depth_prob,
    H,
    W,
):
    batch_size = 2
    dtype = default_floating_dtype()
    key = jax.random.key(43)
    torch.manual_seed(42)
    np.random.seed(42)

    # --- PyTorch Model ---
    torch_block = TorchSwinTransformerBlockV2(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=torch.nn.LayerNorm,
        attn_layer=TorchShiftedWindowAttentionV2,
    )
    torch_block.eval()

    # --- JAX Model ---
    jax_block, state = eqx.nn.make_with_state(SwinTransformerBlockV2)(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=functools.partial(eqx.nn.LayerNorm, eps=1e-05),  # pyright: ignore
        attn_layer=ShiftedWindowAttentionV2,
        inference=False,
        dtype=dtype,
        key=key,
    )
    jax_block, state = eqx.nn.inference_mode((jax_block, state))

    # --- Weight Conversion ---
    weights_dict = torch_block.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    # torchfields = move_running_fields_to_the_end(
    #     torchfields,
    #     identifier="relative_position_index",
    # )
    # torchfields = move_running_fields_to_the_end(
    #     torchfields,
    #     identifier="relative_coords_table",
    # )

    jaxfields, state_indices = pytree_to_fields(
        (
            jax_block,
            state,
        )
    )

    jax_block, state = convert(
        weights_dict,
        (jax_block, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    # --- Inference and Comparison ---
    x_np = np.array(np.random.uniform(size=(batch_size, H, W, dim)), dtype=np.float32)
    x_torch = torch.from_numpy(x_np)
    x_jax = jnp.array(x_np)

    t_out = torch_block(x_torch)
    jax_block_pt = functools.partial(jax_block, key=jax.random.key(2100))
    j_out, state = eqx.filter_vmap(jax_block_pt, in_axes=(0, None), out_axes=(0, None))(
        x_jax, state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-3)


@pytest.mark.skip(
    "If you want to go mad, try to figure out why the fuck this passes locally "
    "but not on ci/cd"
)
def test_swinv1_old():
    torch.set_num_threads(1)
    torch.manual_seed(42)
    np.random.seed(42)

    patch_size = [4, 4]
    embed_dim = 96
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    window_size = [7, 7]
    mlp_ratio = 4.0
    dropout = 0.0
    attention_dropout = 0.0
    stochastic_depth_prob = 0.0
    num_classes = 1000

    torch_swin = TorchSwinTransformer(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        block=TorchSwinTransformerBlock,
    )
    torch_swin.eval()

    x = np.array(np.random.normal(size=(1, 3, 224, 224)), dtype=np.float32)
    x_t = torch.Tensor(x)

    jax_swin, state = eqx.nn.make_with_state(SwinTransformer)(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        norm_layer=None,
        block=SwinTransformerBlock,
        downsample_layer=PatchMerging,
        attn_layer=ShiftedWindowAttention,
        inference=True,
        key=jax.random.key(42),
    )

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)

    jaxfields, state_indices = pytree_to_fields(
        (jax_swin, state),
        model_order=model_orders.get_swin_model_order(1),
    )

    # Debug: print first few jaxfields paths to check ordering
    print("First 5 jaxfield paths:")
    for i, f in enumerate(jaxfields[:5]):
        print(f"  {i}: {f.path}")
    print("First 5 torchfield keys:")
    for i, f in enumerate(torchfields[:5]):
        print(f"  {i}: {f.path}")

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    # Debug: check if head weights match after conversion
    jax_head_w = np.array(jax_swin.head.weight)
    torch_head_w = torch_swin.head.weight.detach().numpy()
    print(
        f"After convert - JAX head.weight shape: {jax_head_w.shape}, first 3: {jax_head_w.flatten()[:3]}"  # noqa
    )
    print(
        f"After convert - Torch head.weight shape: {torch_head_w.shape}, first 3: {torch_head_w.flatten()[:3]}"  # noqa
    )

    # Debug: Check state values (relative_position_index should match torch)
    first_attn = jax_swin.features.layers[1].layers[0].attn
    jax_rel_pos_idx = state.get(first_attn.relative_position_index)
    torch_rel_pos_idx = torch_swin.features[1][0].attn.relative_position_index.numpy()  # noqa  # ty:ignore[not-subscriptable]
    print(
        f"JAX rel_pos_idx shape: {jax_rel_pos_idx.shape}, first 5: {jax_rel_pos_idx.flatten()[:5]}"  # noqa
    )
    print(
        f"Torch rel_pos_idx shape: {torch_rel_pos_idx.shape}, first 5: {torch_rel_pos_idx.flatten()[:5]}"  # noqa
    )
    print(
        f"rel_pos_idx match: {np.allclose(np.array(jax_rel_pos_idx), torch_rel_pos_idx)}"  # noqa
    )

    # Debug: Check relative_position_bias_table
    jax_bias_table = np.array(first_attn.relative_position_bias_table)
    torch_bias_table = (
        torch_swin.features[1][0].attn.relative_position_bias_table.detach().numpy()  # ty:ignore[not-subscriptable]
    )
    print(
        f"JAX bias_table shape: {jax_bias_table.shape}, first 3: {jax_bias_table.flatten()[:3]}"  # noqa
    )
    print(
        f"Torch bias_table shape: {torch_bias_table.shape}, first 3: {torch_bias_table.flatten()[:3]}"  # noqa
    )
    print(
        f"bias_table match: {np.allclose(jax_bias_table, torch_bias_table, atol=1e-6)}"
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))
    key = jax.random.key(22)
    jax_swin_pt = functools.partial(jax_swin, key=key)
    jax_out, state = eqx.filter_vmap(
        jax_swin_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    t_out = torch_swin.forward(x_t).detach().numpy()
    jax_out = np.array(jax_out)

    # Debug: print comparison info
    print(f"JAX out shape: {jax_out.shape}, first 5: {jax_out.flatten()[:5]}")
    print(f"Torch out shape: {t_out.shape}, first 5: {t_out.flatten()[:5]}")
    print(f"Input x first 5: {x.flatten()[:5]}")
    print(f"Max diff: {np.max(np.abs(jax_out - t_out))}")

    assert np.allclose(jax_out, t_out, atol=1e-3)


@pytest.mark.skip(
    "If you want to go mad, try to figure out why the fuck this passes locally "
    "but not on ci/cd"
)
def test_swinv2_old():
    torch.set_num_threads(1)
    torch.manual_seed(42)
    np.random.seed(42)

    patch_size = [4, 4]
    embed_dim = 96
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    window_size = [7, 7]
    mlp_ratio = 4.0
    dropout = 0.0
    attention_dropout = 0.0
    stochastic_depth_prob = 0.0
    num_classes = 1000

    torch_swin = TorchSwinTransformer(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        downsample_layer=TorchPatchMergingV2,
        block=TorchSwinTransformerBlockV2,
    )
    torch_swin.eval()

    x = np.array(np.random.normal(size=(1, 3, 224, 224)), dtype=np.float32)
    x_t = torch.Tensor(x)

    jax_swin, state = eqx.nn.make_with_state(SwinTransformer)(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        norm_layer=None,
        block=SwinTransformerBlockV2,
        downsample_layer=PatchMergingV2,
        attn_layer=ShiftedWindowAttentionV2,
        inference=True,
        key=jax.random.key(42),
    )

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    jaxfields, state_indices = pytree_to_fields(
        (jax_swin, state),
        model_order=model_orders.get_swin_model_order(2),
    )

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )
    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    key = jax.random.key(22)
    jax_swin_pt = functools.partial(jax_swin, key=key)
    jax_out, state = eqx.filter_vmap(
        jax_swin_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    t_out = torch_swin.forward(x_t).detach().numpy()
    jax_out = np.array(jax_out)

    # Debug: print comparison info
    print(f"JAX out shape: {jax_out.shape}, first 5: {jax_out.flatten()[:5]}")
    print(f"Torch out shape: {t_out.shape}, first 5: {t_out.flatten()[:5]}")
    print(f"Input x first 5: {x.flatten()[:5]}")
    print(f"Max diff: {np.max(np.abs(jax_out - t_out))}")

    assert np.allclose(jax_out, t_out, atol=1e-3)
